#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.settings import Settings
from reclass.values.value import Value
from reclass.errors import ResolveError, ParseError
import unittest

SETTINGS = Settings()

def _var(s):
    return '%s%s%s' % (SETTINGS.reference_sentinels[0], s,
                       SETTINGS.reference_sentinels[1])

CONTEXT = {'favcolour':'yellow',
           'motd':{'greeting':'Servus!',
                   'colour':'${favcolour}'
                  },
           'int':1,
           'list':[1,2,3],
           'dict':{1:2,3:4},
           'bool':True
          }

def _poor_mans_template(s, var, value):
    return s.replace(_var(var), value)

class TestValue(unittest.TestCase):

    def test_simple_string(self):
        s = 'my cat likes to hide in boxes'
        tv = Value(s, SETTINGS, '')
        self.assertFalse(tv.has_references)
        self.assertEquals(tv.render(CONTEXT, None), s)

    def _test_solo_ref(self, key):
        s = _var(key)
        tv = Value(s, SETTINGS, '')
        res = tv.render(CONTEXT, None)
        self.assertTrue(tv.has_references)
        self.assertEqual(res, CONTEXT[key])

    def test_solo_ref_string(self):
        self._test_solo_ref('favcolour')

    def test_solo_ref_int(self):
        self._test_solo_ref('int')

    def test_solo_ref_list(self):
        self._test_solo_ref('list')

    def test_solo_ref_dict(self):
        self._test_solo_ref('dict')

    def test_solo_ref_bool(self):
        self._test_solo_ref('bool')

    def test_single_subst_bothends(self):
        s = 'I like ' + _var('favcolour') + ' and I like it'
        tv = Value(s, SETTINGS, '')
        self.assertTrue(tv.has_references)
        self.assertEqual(tv.render(CONTEXT, None),
                         _poor_mans_template(s, 'favcolour',
                                             CONTEXT['favcolour']))

    def test_single_subst_start(self):
        s = _var('favcolour') + ' is my favourite colour'
        tv = Value(s, SETTINGS, '')
        self.assertTrue(tv.has_references)
        self.assertEqual(tv.render(CONTEXT, None),
                         _poor_mans_template(s, 'favcolour',
                                             CONTEXT['favcolour']))

    def test_single_subst_end(self):
        s = 'I like ' + _var('favcolour')
        tv = Value(s, SETTINGS, '')
        self.assertTrue(tv.has_references)
        self.assertEqual(tv.render(CONTEXT, None),
                         _poor_mans_template(s, 'favcolour',
                                             CONTEXT['favcolour']))

    def test_deep_subst_solo(self):
        motd = SETTINGS.delimiter.join(('motd', 'greeting'))
        s = _var(motd)
        tv = Value(s, SETTINGS, '')
        self.assertTrue(tv.has_references)
        self.assertEqual(tv.render(CONTEXT, None),
                         _poor_mans_template(s, motd,
                                             CONTEXT['motd']['greeting']))

    def test_multiple_subst(self):
        greet = SETTINGS.delimiter.join(('motd', 'greeting'))
        s = _var(greet) + ' I like ' + _var('favcolour') + '!'
        tv = Value(s, SETTINGS, '')
        self.assertTrue(tv.has_references)
        want = _poor_mans_template(s, greet, CONTEXT['motd']['greeting'])
        want = _poor_mans_template(want, 'favcolour', CONTEXT['favcolour'])
        self.assertEqual(tv.render(CONTEXT, None), want)

    def test_multiple_subst_flush(self):
        greet = SETTINGS.delimiter.join(('motd', 'greeting'))
        s = _var(greet) + ' I like ' + _var('favcolour')
        tv = Value(s, SETTINGS, '')
        self.assertTrue(tv.has_references)
        want = _poor_mans_template(s, greet, CONTEXT['motd']['greeting'])
        want = _poor_mans_template(want, 'favcolour', CONTEXT['favcolour'])
        self.assertEqual(tv.render(CONTEXT, None), want)

    def test_undefined_variable(self):
        s = _var('no_such_variable')
        tv = Value(s, SETTINGS, '')
        with self.assertRaises(ResolveError):
            tv.render(CONTEXT, None)

    def test_incomplete_variable(self):
        s = SETTINGS.reference_sentinels[0] + 'incomplete'
        with self.assertRaises(ParseError):
            tv = Value(s, SETTINGS, '')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import glob
import itertools
import json
import logging
import os
from collections.abc import Mapping

import yaml
import toml

from kapitan.errors import CompileError, KapitanError
from kapitan.refs.base import Revealer
from kapitan.utils import PrettyDumper

logger = logging.getLogger(__name__)


class InputType(object):
    def __init__(self, type_name, compile_path, search_paths, ref_controller):
        self.type_name = type_name
        self.compile_path = compile_path
        self.search_paths = search_paths
        self.ref_controller = ref_controller

    def compile_obj(self, comp_obj, ext_vars, **kwargs):
        """
        Expand globbed input paths, taking into account provided search paths
        and run compile_input_path() for each resolved input_path.
        kwargs are passed into compile_input_path()
        """
        input_type = comp_obj["input_type"]
        assert input_type == self.type_name

        # expand any globbed paths, taking into account provided search paths
        input_paths = []
        for input_path in comp_obj["input_paths"]:
            globbed_paths = [glob.glob(os.path.join(path, input_path)) for path in self.search_paths]
            inputs = list(itertools.chain.from_iterable(globbed_paths))
            # remove duplicate inputs
            inputs = set(inputs)
            ignore_missing = comp_obj.get("ignore_missing", False)
            if len(inputs) == 0 and not ignore_missing:
                raise CompileError(
                    "Compile error: {} for target: {} not found in "
                    "search_paths: {}".format(input_path, ext_vars["target"], self.search_paths)
                )
            input_paths.extend(inputs)

        for input_path in input_paths:
            self.compile_input_path(input_path, comp_obj, ext_vars, **kwargs)

    def compile_input_path(self, input_path, comp_obj, ext_vars, **kwargs):
        """
        Compile validated input_path in comp_obj
        kwargs are passed into compile_file()
        """
        target_name = ext_vars["target"]
        output_path = comp_obj["output_path"]
        output_type = comp_obj.get("output_type", self.default_output_type())
        prune_input = comp_obj.get("prune", kwargs.get("prune", False))

        logger.debug("Compiling %s", input_path)
        try:
            if kwargs.get("compose_node_name", False):
                _compile_path = os.path.join(self.compile_path, target_name.replace(".", "/"), output_path)
            else:
                _compile_path = os.path.join(self.compile_path, target_name, output_path)
            self.compile_file(
                input_path,
                _compile_path,
                ext_vars,
                output=output_type,
                target_name=target_name,
                prune_input=prune_input,
                **kwargs,
            )
        except KapitanError as e:
            raise CompileError("{}\nCompile error: failed to compile target: {}".format(e, target_name))

    def make_compile_dirs(self, target_name, output_path, **kwargs):
        """make compile dirs, skips if dirs exist"""
        _compile_path = os.path.join(self.compile_path, target_name, output_path)
        if kwargs.get("compose_node_name", False):
            os.makedirs(_compile_path.replace(".", "/"), exist_ok=True)
        else:
            os.makedirs(_compile_path, exist_ok=True)

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """implements compilation for file_path to compile_path with ext_vars"""
        return NotImplementedError

    def default_output_type(self):
        "returns default output_type value"
        return NotImplementedError


class CompilingFile(object):
    def __init__(self, context, fp, ref_controller, **kwargs):
        self.fp = fp
        self.ref_controller = ref_controller
        self.kwargs = kwargs
        self.revealer = Revealer(ref_controller)

    def write(self, data):
        """write data into file"""
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)

        if reveal:
            self.fp.write(self.revealer.reveal_raw(data))
        else:
            self.fp.write(self.revealer.compile_raw(data, target_name=target_name))

    def write_yaml(self, obj):
        """recursively compile or reveal refs and convert obj to yaml and write to file"""
        indent = self.kwargs.get("indent", 2)
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)
        if reveal:
            obj = self.revealer.reveal_obj(obj)
        else:
            obj = self.revealer.compile_obj(obj, target_name=target_name)

        if obj:
            if isinstance(obj, Mapping):
                yaml.dump(
                    obj,
                    stream=self.fp,
                    indent=indent,
                    Dumper=PrettyDumper,
                    default_flow_style=False,
                )
            else:
                yaml.dump_all(
                    obj,
                    stream=self.fp,
                    indent=indent,
                    Dumper=PrettyDumper,
                    default_flow_style=False,
                )

            logger.debug("Wrote %s", self.fp.name)
        else:
            logger.debug("%s is Empty, skipped writing output", self.fp.name)

    def write_json(self, obj):
        """recursively hash or reveal refs and convert obj to json and write to file"""
        indent = self.kwargs.get("indent", 2)
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)
        if reveal:
            obj = self.revealer.reveal_obj(obj)
        else:
            obj = self.revealer.compile_obj(obj, target_name=target_name)
        if obj:
            json.dump(obj, self.fp, indent=indent)
            logger.debug("Wrote %s", self.fp.name)
        else:
            logger.debug("%s is Empty, skipped writing output", self.fp.name)

    def write_toml(self, obj):
        """recursively compile or reveal refs and convert obj to toml and write to file"""
        reveal = self.kwargs.get("reveal", False)
        target_name = self.kwargs.get("target_name", None)
        if reveal:
            obj = self.revealer.reveal_obj(obj)
        else:
            obj = self.revealer.compile_obj(obj, target_name=target_name)
        if obj:
            toml.dump(obj, self.fp)
            logger.debug("Wrote %s", self.fp.name)
        else:
            logger.debug("%s is Empty, skipped writing output", self.fp.name)


class CompiledFile(object):
    def __init__(self, name, ref_controller, **kwargs):
        self.name = name
        self.fp = None
        self.ref_controller = ref_controller
        self.kwargs = kwargs

    def __enter__(self):
        mode = self.kwargs.get("mode", "r")

        # make sure directory for file exists
        os.makedirs(os.path.dirname(self.name), exist_ok=True)

        self.fp = open(self.name, mode)
        return CompilingFile(self, self.fp, self.ref_controller, **self.kwargs)

    def __exit__(self, *args):
        self.fp.close()



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import posix, sys
import traceback

from reclass.defaults import REFERENCE_SENTINELS, EXPORT_SENTINELS
from reclass.utils.dictpath import DictPath

class ReclassException(Exception):

    def __init__(self, rc=posix.EX_SOFTWARE, msg=None, tbFlag=True):
        super(ReclassException, self).__init__()
        self._rc = rc
        self._msg = msg
        if tbFlag:
            self._traceback = traceback.format_exc()
        else:
            self._traceback = None
        self._full_traceback = False

    message = property(lambda self: self._get_message())
    rc = property(lambda self: self._rc)

    def __str__(self):
        return self.message + '\n' + super(ReclassException, self).__str__()

    def _get_message(self):
        if self._msg:
            return self._msg
        else:
            return 'No error message provided.'

    def exit_with_message(self, out=sys.stderr):
        if self._full_traceback:
            t, v, tb = sys.exc_info()
            print('Full Traceback', file=out)
            for l in traceback.format_tb(tb):
                print(l, file=out)
        if self._traceback:
            print(self._traceback, file=out)
        print(self.message, file=out)
        sys.exit(self.rc)


class PermissionError(ReclassException):

    def __init__(self, msg, rc=posix.EX_NOPERM):
        super(PermissionError, self).__init__(rc=rc, msg=msg)


class InvocationError(ReclassException):

    def __init__(self, msg, rc=posix.EX_USAGE):
        super(InvocationError, self).__init__(rc=rc, msg=msg)


class ConfigError(ReclassException):

    def __init__(self, msg, rc=posix.EX_CONFIG):
        super(ConfigError, self).__init__(rc=rc, msg=msg)


class DuplicateUriError(ConfigError):

    def __init__(self, nodes_uri, classes_uri):
        super(DuplicateUriError, self).__init__(msg=None)
        self._nodes_uri = nodes_uri
        self._classes_uri = classes_uri

    def _get_message(self):
        return "The inventory URIs must not be the same " \
               "for nodes and classes: {0}".format(self._nodes_uri)


class UriOverlapError(ConfigError):

    def __init__(self, nodes_uri, classes_uri):
        super(UriOverlapError, self).__init__(msg=None)
        self._nodes_uri = nodes_uri
        self._classes_uri = classes_uri

    def _get_message(self):
        msg = "The URIs for the nodes and classes inventories must not " \
              "overlap, but {0} and {1} do."
        return msg.format(self._nodes_uri, self._classes_uri)


class NotFoundError(ReclassException):

    def __init__(self, msg, rc=posix.EX_IOERR):
        super(NotFoundError, self).__init__(rc=rc, msg=msg)


class NodeNotFound(NotFoundError):

    def __init__(self, storage, nodename, uri):
        super(NodeNotFound, self).__init__(msg=None)
        self.storage = storage
        self.name = nodename
        self.uri = uri

    def _get_message(self):
        msg = "Node '{0}' not found under {1}://{2}"
        return msg.format(self.name, self.storage, self.uri)


class InterpolationError(ReclassException):

    def __init__(self, msg=None, rc=posix.EX_DATAERR, nodename='', uri=None, context=None, tbFlag=True):
        super(InterpolationError, self).__init__(rc=rc, msg=msg, tbFlag=tbFlag)
        self.nodename = nodename
        self.uri = uri
        self.context = context

    def _get_message(self):
        msg = '-> {0}\n'.format(self.nodename)
        msg += self._render_error_message(self._get_error_message(), 1)
        msg = msg[:-1]
        return msg

    def _render_error_message(self, message_list, indent):
        msg = ''
        for l in message_list:
            if isinstance(l, list):
                msg += self._render_error_message(l, indent + 1)
            else:
                msg += (' ' * indent * 3) + l + '\n'
        return msg

    def _add_context_and_uri(self):
        msg = ''
        if self.context:
            msg += ', at %s' % str(self.context)
        if self.uri:
            msg += ', in %s' % self.uri
        return msg


class ClassNotFound(InterpolationError):

    def __init__(self, storage, classname, path, nodename='', uri=None):
        super(ClassNotFound, self).__init__(msg=None, uri=uri, nodename=nodename)
        self.storage = storage
        self.name = classname
        self.path = path

    def _get_error_message(self):
        msg = [ 'In {0}'.format(self.uri),
                'Class {0} not found under {1}://{2}'.format(self.name, self.storage, self.path) ]
        return msg


class ClassNameResolveError(InterpolationError):
    def __init__(self, classname, nodename, uri):
        super(ClassNameResolveError, self).__init__(msg=None, uri=uri, nodename=nodename)
        self.name = classname

    def _get_error_message(self):
        msg = [ 'In {0}'.format(self.uri),
                'Class name {0} not resolvable'.format(self.name) ]
        return msg


class InvQueryClassNotFound(InterpolationError):

    def __init__(self, classNotFoundError, nodename=''):
        super(InvQueryClassNotFound, self).__init__(msg=None, nodename=nodename)
        self.classNotFoundError = classNotFoundError
        self._traceback = self.classNotFoundError._traceback

    def _get_error_message(self):
        msg = [ 'Inventory Queries:',
                '-> {0}'.format(self.classNotFoundError.nodename) ]
        msg.append(self.classNotFoundError._get_error_message())
        return msg


class InvQueryClassNameResolveError(InterpolationError):
    def __init__(self, classNameResolveError, nodename=''):
        super(InvQueryClassNameResolveError, self).__init__(msg=None, nodename=nodename)
        self.classNameResolveError = classNameResolveError
        self._traceback = self.classNameResolveError._traceback

    def _get_error_message(self):
        msg = [ 'Inventory Queries:',
                '-> {0}'.format(self.classNameResolveError.nodename) ]
        msg.append(self.classNameResolveError._get_error_message())
        return msg


class ResolveError(InterpolationError):

    def __init__(self, reference, uri=None, context=None):
        super(ResolveError, self).__init__(msg=None)
        self.reference = reference

    def _get_error_message(self):
        msg = 'Cannot resolve {0}'.format(self.reference.join(REFERENCE_SENTINELS)) + self._add_context_and_uri()
        return [ msg ]


class ResolveErrorList(InterpolationError):
    def __init__(self):
        super(ResolveErrorList, self).__init__(msg=None)
        self.resolve_errors = []
        self._traceback = False

    def add(self, resolve_error):
        self.resolve_errors.append(resolve_error)

    def have_errors(self):
        return len(self.resolve_errors) > 0

    def _get_error_message(self):
        msgs = []
        for e in self.resolve_errors:
            msgs.extend(e._get_error_message())
        return msgs


class InvQueryError(InterpolationError):

    def __init__(self, query, resolveError, uri=None, context=None):
        super(InvQueryError, self).__init__(msg=None)
        self.query = query
        self.resolveError = resolveError
        self._traceback = self.resolveError._traceback

    def _get_error_message(self):
        msg1 = 'Failed inv query {0}'.format(self.query.join(EXPORT_SENTINELS)) + self._add_context_and_uri()
        msg2 = '-> {0}'.format(self.resolveError.nodename)
        msg3 = self.resolveError._get_error_message()
        return [ msg1, msg2, msg3 ]


class ParseError(InterpolationError):

    def __init__(self, msg, line, col, lineno, rc=posix.EX_DATAERR):
        super(ParseError, self).__init__(rc=rc, msg=None)
        self._err = msg
        self._line = line
        self._col = col
        self._lineno = lineno

    def _get_error_message(self):
        msg = [ 'Parse error: {0}'.format(self._line.join(EXPORT_SENTINELS)) + self._add_context_and_uri() ]
        msg.append('{0} at char {1}'.format(self._err, self._col - 1))
        return msg


class InfiniteRecursionError(InterpolationError):

    def __init__(self, context, ref, uri):
        super(InfiniteRecursionError, self).__init__(msg=None, tbFlag=False, uri=uri)
        self.context = context
        self.ref = ref

    def _get_error_message(self):
        msg = [ 'Infinite recursion: {0}'.format(self.ref.join(REFERENCE_SENTINELS)) + self._add_context_and_uri() ]
        return msg


class BadReferencesError(InterpolationError):

    def __init__(self, refs, context, uri):
        super(BadReferencesError, self).__init__(msg=None, context=context, uri=uri, tbFlag=False)
        self.refs = [ r.join(REFERENCE_SENTINELS) for r in refs ]

    def _get_error_message(self):
        msg = [ 'Bad references' + self._add_context_and_uri(),
                '   ' + ', '.join(self.refs) ]
        return msg


class TypeMergeError(InterpolationError):

    def __init__(self, value1, value2, uri):
        super(TypeMergeError, self).__init__(msg=None, uri=uri, tbFlag=False)
        self.type1 = value1.item_type_str()
        self.type2 = value2.item_type_str()

    def _get_error_message(self):
        msg = [ 'Cannot merge {0} over {1}'.format(self.type1, self.type2) + self._add_context_and_uri() ]
        return msg


class ExpressionError(InterpolationError):

    def __init__(self, msg, rc=posix.EX_DATAERR, tbFlag=True):
        super(ExpressionError, self).__init__(rc=rc, msg=None, tbFlag=tbFlag)
        self._error_msg = msg

    def _get_error_message(self):
        msg = [ 'Expression error: {0}'.format(self._error_msg) + self._add_context_and_uri() ]
        return msg


class ChangedConstantError(InterpolationError):

    def __init__(self, uri):
        super(ChangedConstantError, self).__init__(msg=None, uri=uri, tbFlag=False)

    def _get_error_message(self):
        msg = [ 'Attempt to change constant value' + self._add_context_and_uri() ]
        return msg


class MappingError(ReclassException):

    def __init__(self, msg, rc=posix.EX_DATAERR):
        super(MappingError, self).__init__(rc=rc, msg=msg)


class MappingFormatError(MappingError):

    def __init__(self, msg):
        super(MappingFormatError, self).__init__(msg)


class NameError(ReclassException):

    def __init__(self, msg=None, rc=posix.EX_DATAERR):
        super(NameError, self).__init__(rc=rc, msg=msg)


class InvalidClassnameError(NameError):

    def __init__(self, invalid_character, classname):
        super(InvalidClassnameError, self).__init__(msg=None)
        self._char = invalid_character
        self._classname = classname

    def _get_message(self):
        msg = "Invalid character '{0}' in class name '{1}'."
        return msg.format(self._char, self._classname)


class DuplicateNodeNameError(NameError):

    def __init__(self, storage, name, uri1, uri2):
        super(DuplicateNodeNameError, self).__init__(msg=None)
        self._storage = storage
        self._name = name
        self._uris = (uri1, uri2)

    def _get_message(self):
        msg = "{0}: Definition of node '{1}' in '{2}' collides with " \
              "definition in '{3}'. Nodes can only be defined once " \
              "per inventory."
        return msg.format(self._storage, self._name, self._uris[1], self._uris[0])


class MissingModuleError(ReclassException):

    def __init__(self, modname):
        msg = "Module %s is missing" % modname
        super(MissingModuleError, self).__init__(rc=posix.EX_DATAERR, msg=msg)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
from reclass.errors import NotFoundError

SKIPDIRS = ('CVS', 'SCCS')
FILE_EXTENSION = ('.yml', '.yaml')

def vvv(msg):
    #print(msg, file=sys.stderr)
    pass


class Directory(object):

    def __init__(self, path, fileclass=None):
        ''' Initialise a directory object '''
        if not os.path.isdir(path):
            raise NotFoundError('No such directory: %s' % path)
        if not os.access(path, os.R_OK|os.X_OK):
            raise NotFoundError('Cannot change to or read directory: %s' % path)
        self._path = path
        self._fileclass = fileclass
        self._files = {}

    def _register_files(self, dirpath, filenames):
        for f in filter(lambda f: f.endswith(FILE_EXTENSION), filenames):
            vvv('REGISTER {0}'.format(f))
            f = os.path.join(dirpath, f)
            ptr = None if not self._fileclass else self._fileclass(f)
            self._files[f] = ptr

    files = property(lambda self: self._files)

    def walk(self, register_fn=None):
        if not callable(register_fn):
            register_fn = self._register_files

        def _error(exc):
            raise(exc)

        for dirpath, dirnames, filenames in os.walk(self._path,
                                                    topdown=True,
                                                    onerror=_error,
                                                    followlinks=True):
            vvv('RECURSE {0}, {1} files, {2} subdirectories'.format(
                dirpath.replace(os.getcwd(), '.'), len(filenames), len(dirnames)))
            for d in dirnames:
                if d.startswith('.') or d in SKIPDIRS:
                    vvv('   SKIP subdirectory {0}'.format(d))
                    dirnames.remove(d)
            register_fn(dirpath, filenames)

    def __repr__(self):
        return '<{0} {1}>'.format(self.__class__.__name__, self._path)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.output import OutputterBase
import yaml

_SafeDumper = yaml.CSafeDumper if yaml.__with_libyaml__ else yaml.SafeDumper


class Outputter(OutputterBase):

    def dump(self, data, pretty_print=False, no_refs=False):
        if (no_refs):
            return yaml.dump(data, default_flow_style=not pretty_print, Dumper=ExplicitDumper)
        else:
            return yaml.dump(data, default_flow_style=not pretty_print, Dumper=_SafeDumper)


class ExplicitDumper(_SafeDumper):
    """
    A dumper that will never emit aliases.
    """

    def ignore_aliases(self, data):
        return True



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .output import OutputLoader
from .storage.loader import StorageBackendLoader
from .storage.memcache_proxy import MemcacheProxy

def get_storage(storage_type, nodes_uri, classes_uri, compose_node_name, **kwargs):
    storage_class = StorageBackendLoader(storage_type).load()
    return MemcacheProxy(storage_class(nodes_uri, classes_uri, compose_node_name, **kwargs))

def get_path_mangler(storage_type, **kwargs):
    return StorageBackendLoader(storage_type).path_mangler()

def output(data, fmt, pretty_print=False, no_refs=False):
    output_class = OutputLoader(fmt).load()
    outputter = output_class()
    return outputter.dump(data, pretty_print=pretty_print, no_refs=no_refs)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"default values for kapitan variables"

import os

DEFAULT_KUBERNETES_VERSION = "1.14.0"
# standalone is used for standalone json schema (i.e. without external $refs)
# strict is used as it behaves similarly to kubectl
# see https://github.com/garethr/kubernetes-json-schema#kubernetes-json-schemas for more info
SCHEMA_TYPE = "standalone-strict"
FILE_PATH_FORMAT = "v{}-%s/{}.json" % SCHEMA_TYPE

# default path from where user defined custom filters are read
DEFAULT_JINJA2_FILTERS_PATH = os.path.join("lib", "jinja2_filters.py")



from kapitan.inputs import kadet

inv = kadet.inventory()

name = "other_component"
labels = kadet.BaseObj.from_dict({"app": name})


def main():
    output = kadet.BaseObj()
    output.root.my_kadet_component = name
    output.root.labels = labels
    return output



# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import hashlib
import logging
import multiprocessing
import os
from collections import defaultdict, namedtuple
from distutils.dir_util import copy_tree
from functools import partial
from mimetypes import MimeTypes
from shutil import copyfile, rmtree

from git import GitCommandError
from git import Repo
from kapitan.errors import GitSubdirNotFoundError, GitFetchingError, HelmFetchingError
from kapitan.helm_cli import helm_cli
from kapitan.utils import (
    make_request,
    unpack_downloaded_file,
    safe_copy_tree,
    safe_copy_file,
    normalise_join_path,
)

logger = logging.getLogger(__name__)

HelmSource = namedtuple("HelmSource", ["repo", "chart_name", "version", "helm_path"])


def fetch_dependencies(output_path, target_objs, save_dir, force, pool):
    """
    parses through the dependencies parameters in target_objs and fetches them
    all dependencies are first fetched into save_dir, after which they are copied to their respective output_path.
    overwites older version of existing dependencies if force fetced
    """
    # there could be multiple dependency items per source_uri due to reclass inheritance or
    # other user requirements. So create a mapping from source_uri to a set of dependencies with
    # that source_uri
    git_deps = defaultdict(list)
    http_deps = defaultdict(list)
    helm_deps = defaultdict(list)

    # this dict is to make sure no duplicated output_paths exist per source
    deps_output_paths = defaultdict(set)
    for target_obj in target_objs:
        try:
            dependencies = target_obj["dependencies"]
            for item in dependencies:
                dependency_type = item["type"]
                source_uri = item["source"]

                # The target key "output_path" is relative to the compile output path set by the user
                # point to the full output path
                full_output_path = normalise_join_path(output_path, item["output_path"])
                logger.debug("Updated output_path from %s to %s", item["output_path"], output_path)
                item["output_path"] = full_output_path

                if full_output_path in deps_output_paths[source_uri]:
                    # if the output_path is duplicated for the same source_uri
                    logger.warning("Skipping duplicate output path for uri %s", source_uri)
                    continue
                else:
                    deps_output_paths[source_uri].add(full_output_path)

                if dependency_type == "git":
                    git_deps[source_uri].append(item)
                elif dependency_type in ("http", "https"):
                    http_deps[source_uri].append(item)
                elif dependency_type == "helm":
                    version = item.get("version", "")
                    helm_deps[
                        HelmSource(source_uri, item["chart_name"], version, item.get("helm_path"))
                    ].append(item)
                else:
                    logger.warning("%s is not a valid source type", dependency_type)

        except KeyError:
            logger.debug(
                "Target object %s has no 'dependencies' key, continuing", target_obj["vars"]["target"]
            )
            continue

    git_worker = partial(fetch_git_dependency, save_dir=save_dir, force=force)
    http_worker = partial(fetch_http_dependency, save_dir=save_dir, force=force)
    helm_worker = partial(fetch_helm_chart, save_dir=save_dir, force=force)
    [p.get() for p in pool.imap_unordered(http_worker, http_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(git_worker, git_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(helm_worker, helm_deps.items()) if p]


def fetch_git_dependency(dep_mapping, save_dir, force, item_type="Dependency"):
    """
    fetches a git repository at source into save_dir, and copy the repository into
    output_path stored in dep_mapping. ref is used to checkout if exists, fetches master branch by default.
    only subdir is copied into output_path if specified.
    """
    source, deps = dep_mapping
    # to avoid collisions between basename(source)
    path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
    cached_repo_path = os.path.join(save_dir, path_hash + os.path.basename(source))
    if not exists_in_cache(cached_repo_path) or force:
        fetch_git_source(source, cached_repo_path, item_type)
    else:
        logger.debug("Using cached %s %s", item_type, cached_repo_path)
    for dep in deps:
        repo = Repo(cached_repo_path)
        output_path = dep["output_path"]
        copy_src_path = cached_repo_path
        if "ref" in dep:
            ref = dep["ref"]
            repo.git.checkout(ref)
        else:
            repo.git.checkout("master")  # default ref

        if "subdir" in dep:
            sub_dir = dep["subdir"]
            full_subdir = os.path.join(cached_repo_path, sub_dir)
            if os.path.isdir(full_subdir):
                copy_src_path = full_subdir
            else:
                raise GitSubdirNotFoundError(
                    "{} {}: subdir {} not found in repo".format(item_type, source, sub_dir)
                )
        if force:
            copy_tree(copy_src_path, output_path)
        else:
            safe_copy_tree(copy_src_path, output_path)
        logger.info("%s %s: saved to %s", item_type, source, output_path)


def fetch_git_source(source, save_dir, item_type):
    """clones a git repository at source and saves into save_dir"""

    if os.path.exists(save_dir):
        rmtree(save_dir)
        logger.debug("Removed %s", save_dir)
    logger.info("%s %s: fetching now", item_type, source)
    try:
        Repo.clone_from(source, save_dir)
        logger.info("%s %s: successfully fetched", item_type, source)
        logger.debug("Git clone cached to %s", save_dir)
    except GitCommandError as e:
        logger.error(e)
        raise GitFetchingError("{} {}: fetching unsuccessful\n{}".format(item_type, source, e.stderr))


def fetch_http_dependency(dep_mapping, save_dir, force, item_type="Dependency"):
    """
    fetches a http[s] file at source and saves into save_dir, after which it is copied into
    the output_path stored in dep_mapping
    """
    source, deps = dep_mapping
    # to avoid collisions between basename(source)
    path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
    cached_source_path = os.path.join(save_dir, path_hash + os.path.basename(source))
    if not exists_in_cache(cached_source_path) or force:
        content_type = fetch_http_source(source, cached_source_path, item_type)
    else:
        logger.debug("Using cached %s %s", item_type, cached_source_path)
        content_type = MimeTypes().guess_type(cached_source_path)[0]
    for dep in deps:
        output_path = dep["output_path"]
        if dep.get("unpack", False):
            # ensure that the directory we are extracting to exists
            os.makedirs(output_path, exist_ok=True)
            if force:
                is_unpacked = unpack_downloaded_file(cached_source_path, output_path, content_type)
            else:
                unpack_output = os.path.join(
                    save_dir, "extracted-" + str(multiprocessing.current_process().name)
                )
                os.makedirs(unpack_output)
                is_unpacked = unpack_downloaded_file(cached_source_path, unpack_output, content_type)
                safe_copy_tree(unpack_output, output_path)
                # delete unpack output
                rmtree(unpack_output)
            if is_unpacked:
                logger.info("%s %s: extracted to %s", item_type, source, output_path)
            else:
                logger.info(
                    "%s %s: Content-Type %s is not supported for unpack. Ignoring save",
                    item_type,
                    source,
                    content_type,
                )
        else:
            # we are downloading a single file
            parent_dir = os.path.dirname(output_path)
            if parent_dir != "":
                os.makedirs(parent_dir, exist_ok=True)
            if force:
                copyfile(cached_source_path, output_path)
            else:
                safe_copy_file(cached_source_path, output_path)
            logger.info("%s %s: saved to %s", item_type, source, output_path)


def fetch_http_source(source, save_path, item_type):
    """downloads a http[s] file from source and saves into save_path(which is a file)"""

    if os.path.exists(save_path):
        os.remove(save_path)
        logger.debug("Removed %s", save_path)
    logger.info("%s %s: fetching now", item_type, source)
    content, content_type = make_request(source)
    logger.info("%s %s: successfully fetched", item_type, source)
    if content is not None:
        with open(save_path, "wb") as f:
            f.write(content)
        logger.debug("Cached to %s", save_path)
        return content_type
    logger.warning("%s %s: failed to fetch", item_type, source)
    return None


def fetch_helm_chart(dep_mapping, save_dir, force):
    """
    downloads a helm chart and its subcharts from source then untars and moves it to save_dir
    """
    source, deps = dep_mapping

    # to avoid collisions between source.chart_name/source.version
    path_hash = hashlib.sha256(source.repo.encode()).hexdigest()[:8]
    cached_repo_path = os.path.join(
        save_dir, path_hash, source.chart_name + "-" + (source.version or "latest")
    )
    if force or not exists_in_cache(cached_repo_path):
        fetch_helm_archive(source.helm_path, source.repo, source.chart_name, source.version, cached_repo_path)
    else:
        logger.debug("Using cached helm chart at %s", cached_repo_path)

    for dep in deps:
        output_path = dep["output_path"]

        parent_dir = os.path.dirname(output_path)
        if parent_dir != "":
            os.makedirs(parent_dir, exist_ok=True)
        if force:
            copy_tree(cached_repo_path, output_path)
        else:
            safe_copy_tree(cached_repo_path, output_path)
        logger.info("Dependency %s: saved to %s", source.chart_name, output_path)


def fetch_helm_archive(helm_path, repo, chart_name, version, save_path):
    logger.info("Dependency helm chart %s and version %s: fetching now", chart_name, version or "latest")
    # Fetch archive and untar it into parent dir
    save_dir = os.path.dirname(save_path)
    args = ["pull", "--destination", save_dir, "--untar"]

    if version:
        args.append("--version")
        args.append(version)

    if repo.startswith("oci://"):
        args.append(repo)
    else:
        args.append("--repo")
        args.append(repo)
        args.append(chart_name)

    response = helm_cli(helm_path, args)
    if response != "":
        logger.warning("Dependency helm chart %s and version %s: %s", chart_name, version, response)
        raise HelmFetchingError(response)
    else:
        # rename chart to requested name
        os.rename(os.path.join(save_dir, chart_name), save_path)
        logger.info("Dependency helm chart %s and version %s: successfully fetched", chart_name, version)
        logger.info("Dependency helm chart %s and version %s: saved to %s", chart_name, version, save_path)


def exists_in_cache(item_path):
    dep_cache_path = os.path.dirname(item_path)
    if not os.path.exists(dep_cache_path):
        return False
    return os.path.basename(item_path) in os.listdir(dep_cache_path)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.storage import NodeStorageBase

STORAGE_NAME = 'memcache_proxy'

class MemcacheProxy(NodeStorageBase):

    def __init__(self, real_storage, cache_classes=True, cache_nodes=True,
                 cache_nodelist=True):
        name = '{0}({1})'.format(STORAGE_NAME, real_storage.name)
        super(MemcacheProxy, self).__init__(name)
        self._real_storage = real_storage
        self._cache_classes = cache_classes
        if cache_classes:
            self._classes_cache = {}
        self._cache_nodes = cache_nodes
        if cache_nodes:
            self._nodes_cache = {}
        self._cache_nodelist = cache_nodelist
        if cache_nodelist:
            self._nodelist_cache = None

    name = property(lambda self: self._real_storage.name)

    def get_node(self, name, settings):
        if not self._cache_nodes:
            return self._real_storage.get_node(name, settings)
        try:
            return self._nodes_cache[name]
        except KeyError as e:
            ret = self._real_storage.get_node(name, settings)
            self._nodes_cache[name] = ret
        return ret

    def get_class(self, name, environment, settings):
        if not self._cache_classes:
            return self._real_storage.get_class(name, environment, settings)
        try:
            return self._classes_cache[environment][name]
        except KeyError as e:
            if environment not in self._classes_cache:
                self._classes_cache[environment] = dict()
            ret = self._real_storage.get_class(name, environment, settings)
            self._classes_cache[environment][name] = ret
        return ret

    def enumerate_nodes(self):
        if not self._cache_nodelist:
            return self._real_storage.enumerate_nodes()

        elif self._nodelist_cache is None:
            self._nodelist_cache = self._real_storage.enumerate_nodes()

        return self._nodelist_cache



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
import sys

from kapitan.errors import CompileError
from kapitan.inputs.base import CompiledFile, InputType
from kapitan.resources import resource_callbacks, search_imports
from kapitan.utils import prune_empty

logger = logging.getLogger(__name__)


def jsonnet_file(file_path, **kwargs):
    """
    Evaluate file_path jsonnet file.
    kwargs are documented in http://jsonnet.org/implementation/bindings.html
    """
    try:
        if "_jsonnet" not in sys.modules:
            import _jsonnet
        return sys.modules["_jsonnet"].evaluate_file(file_path, **kwargs)
    except ImportError:
        logger.info(
            "Note: Jsonnet is not installed or running on an unsupported architecture."
            " See https://kapitan.dev/compile/#jsonnet"
        )
    except Exception as e:
        raise CompileError(f"Jsonnet error: failed to compile {file_path}:\n {e}")


def go_jsonnet_file(file_path, **kwargs):
    """
    Evaluate file_path jsonnet file using gojsonnet.
    kwargs are documented in http://jsonnet.org/implementation/bindings.html
    """
    try:
        if "_gojsonnet" not in sys.modules:
            import _gojsonnet
        return sys.modules["_gojsonnet"].evaluate_file(file_path, **kwargs)
    except ImportError:
        logger.info(
            "Note: Go-jsonnet is not installed or running on an unsupported architecture."
            " See https://kapitan.dev/compile/#jsonnet"
        )
    except Exception as e:
        raise CompileError(f"Jsonnet error: failed to compile {file_path}:\n {e}")


class Jsonnet(InputType):
    def __init__(self, compile_path, search_paths, ref_controller, use_go=False):
        super().__init__(
            "jsonnet",
            compile_path,
            search_paths,
            ref_controller,
        )
        self.use_go = use_go

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write file_path (jsonnet evaluated) items as files to compile_path.
        ext_vars will be passed as parameters to jsonnet_file()
        kwargs:
            output: default 'yaml', accepts 'json'
            prune: default False, accepts True
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
            indent: default 2
        """

        def _search_imports(cwd, imp):
            return search_imports(cwd, imp, self.search_paths)

        json_output = None
        if self.use_go:
            json_output = go_jsonnet_file(
                file_path,
                import_callback=_search_imports,
                native_callbacks=resource_callbacks(self.search_paths),
                ext_vars=ext_vars,
            )
        else:
            json_output = jsonnet_file(
                file_path,
                import_callback=_search_imports,
                native_callbacks=resource_callbacks(self.search_paths),
                ext_vars=ext_vars,
            )
        output_obj = json.loads(json_output)

        output = kwargs.get("output", "yaml")
        prune = kwargs.get("prune_input", False)
        reveal = kwargs.get("reveal", False)
        target_name = kwargs.get("target_name", None)
        indent = kwargs.get("indent", 2)

        if prune:
            output_obj = prune_empty(output_obj)
            logger.debug("Pruned output for: %s", file_path)

        if not isinstance(output_obj, dict):
            tmp_output_obj = output_obj
            # assume that the output filename is the
            # same as the input jsonnet filename
            filename = os.path.splitext(os.path.basename(file_path))[0]
            output_obj = {}
            output_obj[filename] = tmp_output_obj

        for item_key, item_value in output_obj.items():
            # write each item to disk
            if output == "json":
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_json(item_value)
            elif output in ["yml", "yaml"]:
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_yaml(item_value)
            elif output == "plain":
                file_path = os.path.join(compile_path, "%s" % item_key)
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write(item_value)
            elif output == "toml":
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_toml(item_value)
            else:
                raise ValueError(
                    f"Output type defined in inventory for {file_path} is neither 'json', 'yaml', 'toml' nor 'plain'"
                )

    def default_output_type(self):
        return "yaml"



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"environment refs module"

import base64
import hashlib
import os

from kapitan.refs.base import PlainRef, PlainRefBackend

DEFAULT_ENV_REF_VAR_PREFIX = "KAPITAN_VAR_"


class EnvRef(PlainRef):
    def __init__(self, data, **kwargs):
        """
        writes plain data, which is the "default" value if the ref cannot be located in the KAPITAN_VAR_*
        environment vars prefix during the reveal phase.
        """
        super().__init__(data, kwargs=kwargs)
        self.type_name = "env"

    def reveal(self):
        """
        Attempt to locate the variable in the environment w/ the suffix.
        """
        path_part = self.path.split("/")[-1]
        var_key = "{}{}".format(DEFAULT_ENV_REF_VAR_PREFIX, path_part)
        return os.getenv(var_key, default=os.getenv(var_key.upper(), default=self.data))

    def compile(self):
        """
        Override the way an EnvRef is compiled, since we want to reveal it via the env later.
        """
        compiled = f"?{{{self.type_name}:{self.path}:{self.hash[:8]}}}"
        return compiled


class EnvRefBackend(PlainRefBackend):
    def __init__(self, path, ref_type=EnvRef, **ref_kwargs):
        "Get and create EnvRefs"
        super().__init__(path, ref_type, ref_kwargs=ref_kwargs)
        self.type_name = "env"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–13 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import reclass.cli
reclass.cli.main()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"cached module"

inv = {}
inv_cache = {}
gpg_obj = None
gkms_obj = None
awskms_obj = None
azkms_obj = None
dot_kapitan = {}
ref_controller_obj = None
revealer_obj = None
args = {}  # args won't need resetting
inv_sources = set()


def reset_cache():
    global inv, inv_cache, gpg_obj, gkms_obj, awskms_obj, azkms_obj, dot_kapitan, ref_controller_obj, revealer_obj, inv_sources

    inv = {}
    inv_cache = {}
    inv_sources = set()
    gpg_obj = None
    gkms_obj = None
    awskms_obj = None
    azkms_obj = None
    dot_kapitan = {}
    ref_controller_obj = None
    revealer_obj = None


def from_dict(cache_dict):
    global inv, inv_cache, gpg_obj, gkms_obj, awskms_obj, azkms_obj, dot_kapitan, ref_controller_obj, revealer_obj, inv_sources, args

    inv = cache_dict["inv"]
    inv_cache = cache_dict["inv_cache"]
    inv_sources = cache_dict["inv_sources"]
    gpg_obj = cache_dict["gpg_obj"]
    gkms_obj = cache_dict["gkms_obj"]
    awskms_obj = cache_dict["awskms_obj"]
    azkms_obj = cache_dict["azkms_obj"]
    dot_kapitan = cache_dict["dot_kapitan"]
    ref_controller_obj = cache_dict["ref_controller_obj"]
    revealer_obj = cache_dict["revealer_obj"]
    args = cache_dict["args"]


def as_dict():
    return {
        "inv": inv,
        "inv_cache": inv_cache,
        "inv_sources": inv_sources,
        "gpg_obj": gpg_obj,
        "gkms_obj": gkms_obj,
        "awskms_obj": awskms_obj,
        "azkms_obj": azkms_obj,
        "dot_kapitan": dot_kapitan,
        "ref_controller_obj": ref_controller_obj,
        "revealer_obj": revealer_obj,
        "args": args,
    }


def reset_inv():
    """clears the inv while fetching remote inventories"""
    global inv
    inv = {}



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.storage.yamldata import YamlData

import unittest

class TestYamlData(unittest.TestCase):

    def setUp(self):
        lines = [ 'classes:',
                  '  - testdir.test1',
                  '  - testdir.test2',
                  '  - test3',
                  '',
                  'environment: base',
                  '',
                  'parameters:',
                  '  _TEST_:',
                  '    alpha: 1',
                  '    beta: two' ]
        self.data = '\n'.join(lines)
        self.yamldict = { 'classes': [ 'testdir.test1', 'testdir.test2', 'test3' ],
                          'environment': 'base',
                          'parameters': { '_TEST_': { 'alpha': 1, 'beta': 'two' } }
                        }

    def test_yaml_from_string(self):
        res = YamlData.from_string(self.data, 'testpath')
        self.assertEqual(res.uri, 'testpath')
        self.assertEqual(res.get_data(), self.yamldict)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import contextvars
import inspect
import logging
import os
import sys
from importlib.util import module_from_spec, spec_from_file_location

import kadet
from kadet import BaseModel, BaseObj, Dict

from kapitan import cached
from kapitan.errors import CompileError
from kapitan.inputs.base import CompiledFile, InputType
from kapitan.resources import inventory as inventory_func
from kapitan.utils import prune_empty

# Set external kadet exception to kapitan.error.CompileError
kadet.ABORT_EXCEPTION_TYPE = CompileError

logger = logging.getLogger(__name__)
inventory_path = cached.args.get(
    "inventory_path"
)  # XXX think about this as it probably breaks usage as library
search_paths = contextvars.ContextVar("current search_paths in thread")
current_target = contextvars.ContextVar("current target in thread")


def inventory(lazy=False):
    return Dict(inventory_func(search_paths.get(), current_target.get(), inventory_path), default_box=lazy)


def inventory_global(lazy=False):
    return Dict(inventory_func(search_paths.get(), None, inventory_path), default_box=lazy)


def module_from_path(path, check_name=None):
    """
    loads python module in path
    set check_name to verify module name against spec name
    returns tuple with module and spec
    """

    if not os.path.isdir(path):
        raise FileNotFoundError("path: {} must be an existing directory".format(path))

    module_name = os.path.basename(os.path.normpath(path))
    init_path = os.path.join(path, "__init__.py")
    spec = spec_from_file_location("kadet_component_{}".format(module_name), init_path)
    mod = module_from_spec(spec)

    if spec is None:
        raise ModuleNotFoundError("Could not load module in path {}".format(path))
    if check_name is not None and check_name != module_name:
        raise ModuleNotFoundError(
            "Module name {} does not match check_name {}".format(module_name, check_name)
        )

    return mod, spec


def load_from_search_paths(module_name):
    """
    loads and executes python module with module_name from search paths
    returns module
    """
    for path in search_paths.get():
        try:
            _path = os.path.join(path, module_name)
            mod, spec = module_from_path(_path, check_name=module_name)
            spec.loader.exec_module(mod)
            return mod
        except (ModuleNotFoundError, FileNotFoundError):
            pass
    raise ModuleNotFoundError("Could not load module name {}".format(module_name))


class Kadet(InputType):
    def __init__(self, compile_path, search_paths, ref_controller, input_params={}):
        super().__init__("kadet", compile_path, search_paths, ref_controller)
        self.input_params = input_params

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write file_path (kadet evaluated) items as files to compile_path.
        ext_vars is not used in Kadet
        kwargs:
            output: default 'yaml', accepts 'json'
            prune: default False
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
            indent: default 2
        """
        output = kwargs.get("output", "yaml")
        prune = kwargs.get("prune", False)
        reveal = kwargs.get("reveal", False)
        target_name = kwargs.get("target_name", None)
        # inventory_path = kwargs.get("inventory_path", None)
        indent = kwargs.get("indent", 2)

        current_target.set(target_name)
        search_paths.set(self.search_paths)

        input_params = self.input_params
        # set compile_path allowing kadet functions to have context on where files
        # are being compiled on the current kapitan run
        # we only do this if user didn't pass its own value
        input_params.setdefault("compile_path", compile_path)

        kadet_module, spec = module_from_path(file_path)
        sys.modules[spec.name] = kadet_module
        spec.loader.exec_module(kadet_module)
        logger.debug("Kadet.compile_file: spec.name: %s", spec.name)

        kadet_arg_spec = inspect.getfullargspec(kadet_module.main)
        logger.debug("Kadet main args: %s", kadet_arg_spec.args)

        if len(kadet_arg_spec.args) > 1:
            raise ValueError(f"Kadet {spec.name} main parameters not equal to 1 or 0")

        try:
            if len(kadet_arg_spec.args) == 1:
                output_obj = kadet_module.main(input_params)
            elif len(kadet_arg_spec.args) == 0:
                output_obj = kadet_module.main()
        except Exception:
            # Log traceback and exception as is
            logger.exception("")
            raise CompileError(f"Could not load Kadet module: {spec.name[16:]}")

        output_obj = _to_dict(output_obj)
        if prune:
            output_obj = prune_empty(output_obj)

        # Return None if output_obj has no output
        if not output_obj:
            return None

        for item_key, item_value in output_obj.items():
            # write each item to disk
            if output == "json":
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_json(item_value)
            elif output in ["yml", "yaml"]:
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_yaml(item_value)
            elif output == "plain":
                file_path = os.path.join(compile_path, "%s" % item_key)
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write(item_value)
            elif output == "toml":
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_toml(item_value)
            else:
                raise ValueError(
                    f"Output type defined in inventory for {file_path} is neither 'json', 'yaml', 'toml' nor 'plain'"
                )

    def default_output_type(self):
        return "yaml"


def _to_dict(obj):
    """
    recursively update obj should it contain other
    BaseObj values
    """
    if isinstance(obj, (BaseObj, BaseModel)):
        for k, v in obj.root.items():
            obj.root[k] = _to_dict(v)
        # BaseObj needs to return to_dict()
        return obj.root.to_dict()
    elif isinstance(obj, list):
        obj = [_to_dict(item) for item in obj]
        # list has no .to_dict, return itself
        return obj
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = _to_dict(v)
        # dict has no .to_dict, return itself
        return obj

    # anything else, return itself
    return obj



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–13 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.version import *
from setuptools import setup, find_packages

# use consistent encoding of readme for pypi
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

ADAPTERS = ['salt', 'ansible']
console_scripts = ['reclass = reclass.cli:main']
console_scripts.extend('reclass-{0} = reclass.adapters.{0}:cli'.format(i)
                       for i in ADAPTERS)

setup(
    name = RECLASS_NAME,
    description = DESCRIPTION,
    long_description=long_description,
    version = VERSION,
    author = AUTHOR,
    author_email = AUTHOR_EMAIL,
    maintainer = MAINTAINER,
    maintainer_email = MAINTAINER_EMAIL,
    license = LICENCE,
    url = URL,
    packages = find_packages(exclude=['*tests']), #FIXME validate this
    entry_points = { 'console_scripts': console_scripts },
    install_requires = ['pyparsing', 'pyyaml', 'six', 'enum34', 'ddt'], #FIXME pygit2 (require libffi-dev, libgit2-dev 0.26.x )

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Systems Administration',
        'License :: OSI Approved :: Artistic License',
        'Programming Language :: Python :: 2.7',
    ],

    keywords='enc ansible salt'
)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import sys

import yaml

import reclass.errors
from reclass.datatypes import Entity
from reclass.output.yaml_outputter import ExplicitDumper
from reclass.storage import ExternalNodeStorageBase
from reclass.storage.yamldata import YamlData

from .directory import Directory

FILE_EXTENSION = ('.yml', '.yaml')
STORAGE_NAME = 'yaml_fs'

def vvv(msg):
    #print(msg, file=sys.stderr)
    pass

def path_mangler(inventory_base_uri, nodes_uri, classes_uri):

    if inventory_base_uri is None:
        # if inventory_base is not given, default to current directory
        inventory_base_uri = os.getcwd()

    nodes_uri = nodes_uri or 'nodes'
    classes_uri = classes_uri or 'classes'

    def _path_mangler_inner(path):
        ret = os.path.join(inventory_base_uri, path)
        ret = os.path.expanduser(ret)
        return os.path.abspath(ret)

    n, c = map(_path_mangler_inner, (nodes_uri, classes_uri))
    if n == c:
        raise errors.DuplicateUriError(n, c)
    common = os.path.commonprefix((n, c))
    if common == n or common == c:
        raise errors.UriOverlapError(n, c)

    return n, c


class ExternalNodeStorage(ExternalNodeStorageBase):

    def __init__(self, nodes_uri, classes_uri, compose_node_name):
        super(ExternalNodeStorage, self).__init__(STORAGE_NAME, compose_node_name)

        if nodes_uri is not None:
            self._nodes_uri = nodes_uri
            self._nodes = self._enumerate_inventory(nodes_uri, self.node_name_mangler)

        if classes_uri is not None:
            self._classes_uri = classes_uri
            self._classes = self._enumerate_inventory(classes_uri, self.class_name_mangler)

    nodes_uri = property(lambda self: self._nodes_uri)
    classes_uri = property(lambda self: self._classes_uri)

    def _enumerate_inventory(self, basedir, name_mangler):
        ret = {}
        def register_fn(dirpath, filenames):
            filenames = [f for f in filenames if f.endswith(FILE_EXTENSION)]
            vvv('REGISTER {0} in path {1}'.format(filenames, dirpath))
            for f in filenames:
                name = os.path.splitext(f)[0]
                relpath = os.path.relpath(dirpath, basedir)
                if callable(name_mangler):
                    relpath, name = name_mangler(relpath, name)
                uri = os.path.join(dirpath, f)
                if name in ret:
                    E = reclass.errors.DuplicateNodeNameError
                    raise E(self.name, name,
                            os.path.join(basedir, ret[name]), uri)
                if relpath:
                    f = os.path.join(relpath, f)
                ret[name] = f

        d = Directory(basedir)
        d.walk(register_fn)
        return ret

    def get_node(self, name, settings):
        vvv('GET NODE {0}'.format(name))
        try:
            relpath = self._nodes[name]
            path = os.path.join(self.nodes_uri, relpath)
            pathname = os.path.splitext(relpath)[0]
        except KeyError as e:
            raise reclass.errors.NodeNotFound(self.name, name, self.nodes_uri)
        entity = YamlData.from_file(path).get_entity(name, pathname, settings)
        return entity

    def get_class(self, name, environment, settings):
        vvv('GET CLASS {0}'.format(name))
        try:
            path = os.path.join(self.classes_uri, self._classes[name])
            pathname = os.path.splitext(self._classes[name])[0]
        except KeyError as e:
            raise reclass.errors.ClassNotFound(self.name, name, self.classes_uri)
        entity = YamlData.from_file(path).get_entity(name, pathname, settings)
        return entity

    def enumerate_nodes(self):
        return self._nodes.keys()



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"reference functions"

import base64
import hashlib
import logging
import secrets  # python secrets module
import string

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, rsa
from kapitan.errors import RefError

logger = logging.getLogger(__name__)


def eval_func(func_name, ctx, *func_params):
    """calls specific function which generates the secret"""
    func_lookup = get_func_lookup()

    return func_lookup[func_name](ctx, *func_params)


def get_func_lookup():
    """returns the lookup-table for the generator functions"""
    return {
        "randomstr": randomstr,
        "random": random,
        "sha256": sha256,
        "ed25519": ed25519_private_key,
        "rsa": rsa_private_key,
        "rsapublic": rsa_public_key,
        "publickey": public_key,
        "reveal": reveal,
        "loweralphanum": loweralphanum,
        "basicauth": basicauth,
    }


def randomstr(ctx, nbytes=""):
    """
    generates a URL-safe text string, containing nbytes random bytes
    sets it to ctx.data
    """
    # deprecated function
    logger.info("DeprecationWarning: randomstr is deprecated. Use random:str instead")
    random(ctx, "str", nbytes)


def sha256(ctx, salt=""):
    """sets ctx.data to salted sha256 hexdigest for input_value"""
    if ctx.data:
        salted_input_value = salt + ":" + ctx.data
        ctx.data = hashlib.sha256(salted_input_value.encode()).hexdigest()
    else:
        raise RefError(
            "Ref error: eval_func: nothing to sha256 hash; try " "something like '|random:str|sha256'"
        )


def ed25519_private_key(ctx):
    """sets ctx.data to a ed25519 private key"""

    key = ed25519.Ed25519PrivateKey.generate()

    ctx.data = str(
        key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ),
        "utf-8",
    )


def rsa_private_key(ctx, key_size="4096"):
    """sets ctx.data to a RSA private key of key_size, default 4096"""
    rsa_key_size = int(key_size)

    key = rsa.generate_private_key(public_exponent=65537, key_size=rsa_key_size, backend=default_backend())

    ctx.data = str(
        key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ),
        "utf-8",
    )


def rsa_public_key(ctx):
    """Derives RSA public key from revealed private key"""
    if not ctx.data:
        raise RefError(
            "Ref error: eval_func: RSA public key cannot be derived; try "
            "something like '|reveal:path/to/encrypted_private_key|rsapublic'"
        )

    public_key(ctx)


def public_key(ctx):
    """Derives RSA public key from revealed private key"""
    if not ctx.data:
        raise RefError(
            "Ref error: eval_func: public key cannot be derived; try "
            "something like '|reveal:path/to/encrypted_private_key|publickey'"
        )

    data_dec = ctx.data
    if ctx.ref_encoding == "base64":
        data_dec = base64.b64decode(data_dec).decode()

    private_key = serialization.load_pem_private_key(
        data_dec.encode(), password=None, backend=default_backend()
    )
    public_key = private_key.public_key()

    ctx.data = str(
        public_key.public_bytes(
            encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
        ),
        "UTF-8",
    )


def reveal(ctx, secret_path):
    """
    decrypt and return data from secret_path
    """
    token_type_name = ctx.ref_controller.token_type_name(ctx.token)
    secret_tag = "?{{{}:{}}}".format(token_type_name, secret_path)
    try:
        ref_obj = ctx.ref_controller[secret_tag]
        ctx.ref_encoding = ref_obj.encoding
        ctx.data = ref_obj.reveal()
    except KeyError:
        raise RefError(
            f"|reveal function error: {secret_path} file in {ctx.token}|reveal:{secret_path} does not exist"
        )


def loweralphanum(ctx, nchars="8"):
    """generates a DNS-compliant text string (a-z and 0-9), containing lower alphanum chars"""
    # deprecated function
    logger.info("DeprecationWarning: loweralphanum is deprecated. Use random:loweralphanum instead")
    random(ctx, "loweralphanum", nchars)


def random(ctx, type="str", nchars="", special_chars=string.punctuation):
    """
    generates a text string, containing nchars of given type
    """

    pool_lookup = {
        "str": string.ascii_letters + string.digits + "-_",
        "int": string.digits,
        "loweralpha": string.ascii_lowercase,
        "upperalpha": string.ascii_uppercase,
        "loweralphanum": string.ascii_lowercase + string.digits,
        "upperalphanum": string.ascii_uppercase + string.digits,
        "special": string.ascii_letters + string.digits + special_chars,
    }

    default_nchars_lookup = {
        "str": 43,
        "int": 16,
    }

    # get pool of given type
    pool = pool_lookup.get(type, None)
    if not pool:
        raise RefError(
            "{}: unknown random type used. Choose one of {}".format(type, [key for key in pool_lookup])
        )

    # get default value for nchars if nchars is not specified
    if not nchars:
        nchars = default_nchars_lookup.get(type, 8)
    else:
        # check input for nchars
        try:
            nchars = int(nchars)
        except ValueError:
            raise RefError(f"Ref error: eval_func: {nchars} cannot be converted into integer.")

    # check if any special characters are specified without using type special
    if type != "special" and special_chars != string.punctuation:
        raise RefError(
            "Ref error: eval_func: {} has no option to use special characters. Use type special instead, i.e. ||random:special:{}".format(
                type, special_chars
            )
        )

    # check if pool is valid, eliminates duplicates
    allowed_pool = string.ascii_letters + string.digits + string.punctuation
    pool = "".join(set(pool).intersection(allowed_pool))

    # generate string based on given pool
    generated_str = "".join(secrets.choice(pool) for i in range(nchars))

    # set ctx.data to generated string
    ctx.data = generated_str


def basicauth(ctx, username="", password=""):
    # check if parameters are specified
    if not username:
        # use random pet name as username
        username = "".join(secrets.choice(string.ascii_lowercase) for i in range(8))

    if not password:
        # generate random password
        pool = string.ascii_letters + string.digits
        password = "".join(secrets.choice(pool) for i in range(8))
    # generate basic-auth token (base64-encoded)
    token = username + ":" + password
    token_bytes = token.encode()
    token_bytes_b64 = base64.b64encode(token_bytes)
    token_b64 = token_bytes_b64.decode()

    # set generated token to ctx.data
    ctx.data = token_b64



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import shutil
from distutils.dir_util import copy_tree

from kapitan.inputs.base import InputType

logger = logging.getLogger(__name__)


class Copy(InputType):
    def __init__(self, compile_path, search_paths, ref_controller, ignore_missing=False):
        self.ignore_missing = ignore_missing
        super().__init__("copy", compile_path, search_paths, ref_controller)

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write items in path as plain rendered files to compile_path.
        path can be either a file or directory.
        """

        # Whether to fail silently if the path does not exists.
        ignore_missing = self.ignore_missing
        try:
            logger.debug("Copying %s to %s.", file_path, compile_path)
            if os.path.exists(file_path):
                if os.path.isfile(file_path):
                    if os.path.isfile(compile_path):
                        shutil.copy2(file_path, compile_path)
                    else:
                        os.makedirs(compile_path, exist_ok=True)
                        shutil.copy2(file_path, os.path.join(compile_path, os.path.basename(file_path)))
                else:
                    compile_path = os.path.abspath(compile_path)  # Resolve relative paths
                    copy_tree(file_path, compile_path)
            elif ignore_missing == False:
                raise OSError(f"Path {file_path} does not exist and `ignore_missing` is {ignore_missing}")
        except OSError as e:
            logger.exception("Input dir not copied. Error: %s", e)

    def default_output_type(self):
        # no output_type options for copy
        return None



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .parser import Parser
from .dictitem import DictItem
from .listitem import ListItem
from .scaitem import ScaItem
from reclass.errors import InterpolationError

from six import string_types

class Value(object):

    _parser = Parser()

    def __init__(self, value, settings, uri, parse_string=True):
        self._settings = settings
        self.uri = uri
        self.overwrite = False
        self.constant = False
        if isinstance(value, string_types):
            if parse_string:
                try:
                    self._item = self._parser.parse(value, self._settings)
                except InterpolationError as e:
                    e.uri = self.uri
                    raise
            else:
                self._item = ScaItem(value, self._settings)
        elif isinstance(value, list):
            self._item = ListItem(value, self._settings)
        elif isinstance(value, dict):
            self._item = DictItem(value, self._settings)
        else:
            self._item = ScaItem(value, self._settings)

    def item_type(self):
        return self._item.type

    def item_type_str(self):
        return self._item.type_str()

    def is_container(self):
        return self._item.is_container()

    @property
    def allRefs(self):
        return self._item.allRefs

    @property
    def has_references(self):
        return self._item.has_references

    @property
    def has_inv_query(self):
        return self._item.has_inv_query

    @property
    def needs_all_envs(self):
        if self._item.has_inv_query:
            return self._item.needs_all_envs
        return False

    def ignore_failed_render(self):
        return self._item.ignore_failed_render

    @property
    def is_complex(self):
        return self._item.is_complex

    def get_references(self):
        return self._item.get_references()

    def get_inv_references(self):
        return self._item.get_inv_references()

    def assembleRefs(self, context):
        if self._item.has_references:
            self._item.assembleRefs(context)

    def render(self, context, inventory):
        try:
            return self._item.render(context, inventory)
        except InterpolationError as e:
            e.uri = self.uri
            raise

    @property
    def contents(self):
        return self._item.contents

    def merge_over(self, value):
        self._item = self._item.merge_over(value._item)
        return self

    def __repr__(self):
        return 'Value(%r)' % self._item

    def __str__(self):
        return str(self._item)



"random utils"
# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import print_function

import collections
import json
import logging
import magic
import math
import os
import re
import stat
import sys
import tarfile
import traceback
from collections import Counter, defaultdict
from distutils.dir_util import mkpath
from distutils.errors import DistutilsFileError
from distutils.file_util import _copy_file_contents
from functools import lru_cache, wraps
from hashlib import sha256
from zipfile import ZipFile

import jinja2
import requests
import yaml

from kapitan import cached, defaults
from kapitan.errors import CompileError
from kapitan.inputs.jinja2_filters import load_jinja2_filters, load_jinja2_filters_from_file
from kapitan.version import VERSION

logger = logging.getLogger(__name__)


try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader


def fatal_error(message):
    "Logs error message, sys.exit(1)"
    logger.error(message)
    sys.exit(1)


def hashable_lru_cache(func):
    """Usable instead of lru_cache for functions using unhashable objects"""

    cache = lru_cache(maxsize=256)

    def deserialise(value):
        try:
            return json.loads(value)
        except Exception:
            logger.debug("hashable_lru_cache: %s not serialiseable, using generic lru_cache instead", value)
            return value

    def func_with_serialized_params(*args, **kwargs):
        _args = tuple([deserialise(arg) for arg in args])
        _kwargs = {k: deserialise(v) for k, v in kwargs.items()}
        return func(*_args, **_kwargs)

    cached_function = cache(func_with_serialized_params)

    @wraps(func)
    def lru_decorator(*args, **kwargs):
        _args = tuple([json.dumps(arg, sort_keys=True) if type(arg) in (list, dict) else arg for arg in args])
        _kwargs = {
            k: json.dumps(v, sort_keys=True) if type(v) in (list, dict) else v for k, v in kwargs.items()
        }
        return cached_function(*_args, **_kwargs)

    lru_decorator.cache_info = cached_function.cache_info
    lru_decorator.cache_clear = cached_function.cache_clear
    return lru_decorator


class termcolor:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def normalise_join_path(dirname, path):
    """Join dirname with path and return in normalised form"""
    logger.debug(os.path.normpath(os.path.join(dirname, path)))
    return os.path.normpath(os.path.join(dirname, path))


@lru_cache(maxsize=256)
def render_jinja2_template(content, context):
    """Render jinja2 content with context"""
    return jinja2.Template(content, undefined=jinja2.StrictUndefined).render(context)


@lru_cache(maxsize=256)
def sha256_string(string):
    """Returns sha256 hex digest for string"""
    return sha256(string.encode("UTF-8")).hexdigest()


def _jinja_error_info(trace_data):
    """Extract jinja2 templating related frames from traceback data"""
    try:
        return [x for x in trace_data if x[2] in ("top-level template code", "template", "<module>")][-1]
    except IndexError:
        pass


def render_jinja2_file(name, context, jinja2_filters=defaults.DEFAULT_JINJA2_FILTERS_PATH, search_paths=None):
    """Render jinja2 file name with context"""
    path, filename = os.path.split(name)
    search_paths = [path or "./"] + (search_paths or [])
    env = jinja2.Environment(
        undefined=jinja2.StrictUndefined,
        loader=jinja2.FileSystemLoader(search_paths),
        trim_blocks=True,
        lstrip_blocks=True,
        extensions=["jinja2.ext.do"],
    )
    load_jinja2_filters(env)
    load_jinja2_filters_from_file(env, jinja2_filters)
    try:
        return env.get_template(filename).render(context)
    except jinja2.TemplateError as e:
        # Exception misses the line number info. Retreive it from traceback
        err_info = _jinja_error_info(traceback.extract_tb(sys.exc_info()[2]))
        raise CompileError(f"Jinja2 TemplateError: {e}, at {err_info[0]}:{err_info[1]}")


def render_jinja2(path, context, jinja2_filters=defaults.DEFAULT_JINJA2_FILTERS_PATH, search_paths=None):
    """
    Render files in path with context
    Returns a dict where the is key is the filename (with subpath)
    and value is a dict with content and mode
    Empty paths will not be rendered
    Path can be a single file or directory
    Ignores hidden files (.filename)
    """
    rendered = {}
    walk_root_files = []
    if os.path.isfile(path):
        dirname = os.path.dirname(path)
        basename = os.path.basename(path)
        walk_root_files = [(dirname, None, [basename])]
    else:
        walk_root_files = os.walk(path)

    for root, _, files in walk_root_files:
        for f in files:
            if f.startswith("."):
                logger.debug("render_jinja2: ignoring file %s", f)
                continue
            render_path = os.path.join(root, f)
            logger.debug("render_jinja2 rendering %s", render_path)
            # get subpath and filename, strip any leading/trailing /
            name = render_path[len(os.path.commonprefix([root, path])) :].strip("/")
            try:
                rendered[name] = {
                    "content": render_jinja2_file(
                        render_path, context, jinja2_filters=jinja2_filters, search_paths=search_paths
                    ),
                    "mode": file_mode(render_path),
                }
            except Exception as e:
                raise CompileError(f"Jinja2 error: failed to render {render_path}: {e}")

    return rendered


def file_mode(name):
    """Returns mode for file name"""
    st = os.stat(name)
    return stat.S_IMODE(st.st_mode)


def prune_empty(d):
    """
    Remove empty lists and empty dictionaries from d
    (similar to jsonnet std.prune but faster)
    """
    if not isinstance(d, (dict, list)):
        return d

    if isinstance(d, list):
        if len(d) > 0:
            return [v for v in (prune_empty(v) for v in d) if v is not None]

    if isinstance(d, dict):
        if len(d) > 0:
            return {k: v for k, v in ((k, prune_empty(v)) for k, v in d.items()) if v is not None}


class PrettyDumper(yaml.SafeDumper):
    """
    Increases indent of nested lists.
    By default, they are indendented at the same level as the key on the previous line
    More info on https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation
    """

    def increase_indent(self, flow=False, indentless=False):
        return super(PrettyDumper, self).increase_indent(flow, False)


def multiline_str_presenter(dumper, data):
    """
    Configures yaml for dumping multiline strings with given style.
    By default, strings are getting dumped with style='"'.
    Ref: https://github.com/yaml/pyyaml/issues/240#issuecomment-1018712495
    """
    # get parsed args from cached.py
    compile_args = cached.args.get("compile", None)
    style = None
    if compile_args:
        style = compile_args.yaml_multiline_string_style

    # check for inventory args too
    inventory_args = cached.args.get("inventory", None)
    if inventory_args:
        style = inventory_args.multiline_string_style

    if style == "literal":
        style = "|"
    elif style == "folded":
        style = ">"
    else:
        style = '"'
    if data.count("\n") > 0:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=style)
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


PrettyDumper.add_representer(str, multiline_str_presenter)


def null_presenter(dumper, data):
    """Configures yaml for omitting value from null-datatype"""
    # get parsed args from cached.py
    compile_args = cached.args.get("compile", None)
    flag_value = None
    if compile_args:
        flag_value = compile_args.yaml_dump_null_as_empty

    if flag_value:
        return dumper.represent_scalar("tag:yaml.org,2002:null", "")
    else:
        return dumper.represent_scalar("tag:yaml.org,2002:null", "null")


PrettyDumper.add_representer(type(None), null_presenter)


def flatten_dict(d, parent_key="", sep="."):
    """Flatten nested elements in a dictionary"""
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.abc.MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


@hashable_lru_cache
def deep_get(dictionary, keys, previousKey=None):
    """Search recursively for 'keys' in 'dictionary' and return value, otherwise return None"""
    value = None
    if len(keys) > 0:
        value = dictionary.get(keys[0], None) if isinstance(dictionary, dict) else None

        if value:
            # If we are at the last key and we have a value, we are done
            if len(keys) == 1:
                return value

            # If we have more variables in the search chain and we don't have a dict, return not found
            if not isinstance(value, dict):
                return None

            # Recurse with next keys in the chain on the dict
            return deep_get(value, keys[1:], previousKey=keys[0])
        else:
            if isinstance(dictionary, dict):
                # If we find nothing, check for globbing, loop and match with dict keys
                if "*" in keys[0]:
                    key_lower = keys[0].replace("*", "").lower()
                    for dict_key in dictionary.keys():
                        if key_lower in dict_key.lower():
                            # If we're at the last key in the chain, return matched value
                            if len(keys) == 1:
                                return dictionary[dict_key]

                            # If we have more variables in the chain, continue recursion
                            return deep_get(dictionary[dict_key], keys[1:], previousKey=keys[0])

                if not previousKey:
                    # No previous keys in chain and no globbing, move down the dictionary and recurse
                    for v in dictionary.values():
                        if isinstance(v, dict):
                            item = None
                            if len(keys) > 1:
                                item = deep_get(v, keys, previousKey=keys[0])
                            else:
                                item = deep_get(v, keys)

                            if item is not None:
                                return item

    return value


def searchvar(args):
    """Show all inventory files where a given reclass variable is declared"""
    output = []
    maxlength = 0
    keys = args.searchvar.split(".")
    for full_path in list_all_paths(args.inventory_path):
        if full_path.endswith(".yml") or full_path.endswith(".yaml"):
            with open(full_path, "r") as fd:
                data = yaml.load(fd, Loader=YamlLoader)
                value = deep_get(data, keys)
                if value is not None:
                    output.append((full_path, value))
                    if len(full_path) > maxlength:
                        maxlength = len(full_path)
    if args.pretty_print:
        for i in output:
            print(i[0])
            for line in yaml.dump(i[1], default_flow_style=False).splitlines():
                print("    ", line)
            print()
    else:
        for i in output:
            print("{0!s:{length}} {1!s}".format(*i, length=maxlength + 2))


def directory_hash(directory):
    """Return the sha256 hash for the file contents of a directory"""
    if not os.path.exists(directory):
        raise IOError(f"utils.directory_hash failed, {directory} dir doesn't exist")

    if not os.path.isdir(directory):
        raise IOError(f"utils.directory_hash failed, {directory} is not a directory")

    try:
        hash = sha256()
        for root, _, files in sorted(os.walk(directory)):
            for names in sorted(files):
                file_path = os.path.join(root, names)
                try:
                    with open(file_path, "r") as f:
                        file_hash = sha256(f.read().encode("UTF-8"))
                        hash.update(file_hash.hexdigest().encode("UTF-8"))
                except Exception as e:
                    if isinstance(e, UnicodeDecodeError):
                        with open(file_path, "rb") as f:
                            binary_file_hash = sha256(f.read())
                            hash.update(binary_file_hash.hexdigest().encode("UTF-8"))
                    else:
                        raise CompileError(f"utils.directory_hash failed to open {file_path}: {e}")
    except Exception as e:
        raise CompileError(f"utils.directory_hash failed: {e}")

    return hash.hexdigest()


def dictionary_hash(dict):
    """Return the sha256 hash for dict"""
    return sha256(json.dumps(dict, sort_keys=True).encode("UTF-8")).hexdigest()


def get_entropy(s):
    """Computes and returns the Shannon Entropy for string 's'"""
    length = float(len(s))
    # https://en.wiktionary.org/wiki/Shannon_entropy
    entropy = -sum(count / length * math.log(count / length, 2) for count in Counter(s).values())
    return round(entropy, 2)


def list_all_paths(folder):
    """Given a folder (string), returns a list with the full paths
    of every sub-folder/file.
    """
    for root, folders, files in os.walk(folder):
        for filename in folders + files:
            yield os.path.join(root, filename)


def dot_kapitan_config():
    """Returns the parsed YAML .kapitan file. Subsequent requests will be cached"""
    if not cached.dot_kapitan:
        if os.path.exists(".kapitan"):
            with open(".kapitan", "r") as f:
                cached.dot_kapitan = yaml.safe_load(f)

    return cached.dot_kapitan


def from_dot_kapitan(command, flag, default):
    """
    Returns the 'flag' for 'command' from .kapitan file. If failed, returns 'default'
    """
    kapitan_config = dot_kapitan_config()

    try:
        if kapitan_config[command]:
            flag_value = kapitan_config[command][flag]
            if flag_value:
                return flag_value
    except KeyError:
        pass

    return default


def compare_versions(v1_raw, v2_raw):
    """
    Parses v1_raw and v2_raw into versions and compares them
    Returns 'equal' if v1 == v2
    Returns 'greater' if v1 > v2
    Returns 'lower' if v1 < v2
    """
    v1 = v1_raw.replace("-rc", "")
    v2 = v2_raw.replace("-rc", "")
    v1_split = v1.split(".")
    v2_split = v2.split(".")
    min_range = min(len(v1_split), len(v2_split))

    for i in range(min_range):
        if v1_split[i] == v2_split[i]:
            continue
        if v1_split[i] > v2_split[i]:
            return "greater"
        if v1_split[i] < v2_split[i]:
            return "lower"

    if min_range > 2:
        v1_is_rc = "-rc" in v1_raw
        v2_is_rc = "-rc" in v2_raw

        if not v1_is_rc and v2_is_rc:
            return "greater"
        elif v1_is_rc and not v2_is_rc:
            return "lower"

    return "equal"


def check_version():
    """
    Checks the version in .kapitan is the same as the current version.
    If the version in .kapitan is greater, it will prompt to upgrade.
    If the version in .kapitan is lower, it will prompt to update .kapitan or downgrade.
    """
    kapitan_config = dot_kapitan_config()
    try:
        if kapitan_config and kapitan_config["version"]:
            dot_kapitan_version = str(kapitan_config["version"])
            result = compare_versions(dot_kapitan_version, VERSION)
            if result == "equal":
                return
            print(f"{termcolor.WARNING}Current version: {VERSION}")
            print(f"Version in .kapitan: {dot_kapitan_version}{termcolor.ENDC}\n")

            # If .kapitan version is greater than current version
            if result == "greater":
                print(f"Upgrade kapitan to '{dot_kapitan_version}' in order to keep results consistent:\n")
            # If .kapitan version is lower than current version
            elif result == "lower":
                print(f"Option 1: You can update the version in .kapitan to '{VERSION}' and recompile\n")
                print(
                    f"Option 2: Downgrade kapitan to '{dot_kapitan_version}' in order to keep results consistent:\n"
                )

            print(f"Docker: docker pull kapicorp/kapitan:{dot_kapitan_version}")
            print(f"Pip (user): pip3 install --user --upgrade kapitan=={dot_kapitan_version}\n")
            print("Check https://github.com/kapicorp/kapitan#quickstart for more info.\n")
            print(
                "If you know what you're doing, you can skip this check by adding '--ignore-version-check'."
            )
            sys.exit(1)
    except KeyError:
        pass


def search_target_token_paths(target_secrets_path, targets):
    """
    returns dict of target and their secret token paths (e.g ?{[gpg/gkms/awskms]:path/to/secret}) in target_secrets_path
    targets is a set of target names used to lookup targets in target_secrets_path
    directory should be structured as follow ./refs/${target_name}/file
    """
    target_files = defaultdict(list)
    for full_path in list_all_paths(target_secrets_path):
        secret_path = full_path[len(target_secrets_path) + 1 :]
        target_name = secret_path.split("/")[0]
        if target_name in targets and os.path.isfile(full_path):
            with open(full_path) as fp:
                obj = yaml.load(fp, Loader=YamlLoader)
                try:
                    secret_type = obj["type"]
                except KeyError:
                    # Backwards compatible with gpg secrets that didn't have type in yaml
                    secret_type = "gpg"
                secret_path = f"?{{{secret_type}:{secret_path}}}"
            logger.debug("search_target_token_paths: found %s", secret_path)
            target_files[target_name].append(secret_path)
    return target_files


def make_request(source):
    """downloads the http file at source and returns it's content"""
    r = requests.get(source)
    if r.ok:
        return r.content, r.headers["Content-Type"]
    else:
        r.raise_for_status()
    return None, None


def unpack_downloaded_file(file_path, output_path, content_type):
    """unpacks files of various MIME type and stores it to the output_path"""

    if content_type == None or content_type == "application/octet-stream":
        if re.search(r"^Zip archive data.*", magic.from_file(file_path)):
            content_type = "application/zip"

    if content_type == "application/x-tar":
        tar = tarfile.open(file_path)
        tar.extractall(path=output_path)
        tar.close()
        is_unpacked = True
    elif content_type == "application/zip":
        zfile = ZipFile(file_path)
        zfile.extractall(output_path)
        zfile.close()
        is_unpacked = True
    elif content_type in [
        "application/gzip",
        "application/octet-stream",
        "application/x-gzip",
        "application/x-compressed",
        "application/x-compressed-tar",
    ]:
        if re.search(r"(\.tar\.gz|\.tgz)$", file_path):
            tar = tarfile.open(file_path)
            tar.extractall(path=output_path)
            tar.close()
            is_unpacked = True
        else:
            extension = re.findall(r"\..*$", file_path)[0]
            logger.debug("File extension %s not suported", extension)
            is_unpacked = False
    else:
        logger.debug("Content type %s not suported", content_type)
        is_unpacked = False
    return is_unpacked


def safe_copy_file(src, dst):
    """Copy a file from 'src' to 'dst'.

    Similar to distutils.file_util.copy_file except
    if the file exists in 'dst' it's not clobbered
    or overwritten.

    returns a tupple (src, val)
    file not copied if val = 0 else 1
    """

    if not os.path.isfile(src):
        raise DistutilsFileError("Can't copy {}: doesn't exist or is not a regular file".format(src))

    if os.path.isdir(dst):
        dir = dst
        dst = os.path.join(dst, os.path.basename(src))
    else:
        dir = os.path.dirname(dst)

    if os.path.isfile(dst):
        logger.warning("Not updating %s (file already exists)", dst)
        return (dst, 0)
    _copy_file_contents(src, dst)
    logger.debug("Copied %s to %s", src, dir)
    return (dst, 1)


def safe_copy_tree(src, dst):
    """Recursively copies the 'src' directory tree to 'dst'

    Both 'src' and 'dst' must be directories.
    similar to distutil.dir_util.copy_tree except
    it doesn't overwite an existing file and doesn't
    copy any file starting with "."

    Returns a list of copied file paths.
    """
    if not os.path.isdir(src):
        raise DistutilsFileError("Cannot copy tree {}: not a directory".format(src))
    try:
        names = os.listdir(src)
    except OSError as e:
        raise DistutilsFileError("Error listing files in {}: {}".format(src, e.strerror))

    mkpath(dst)
    outputs = []

    for name in names:
        src_name = os.path.join(src, name)
        dst_name = os.path.join(dst, name)

        if name.startswith("."):
            logger.debug("Not copying %s", src_name)
            continue
        if os.path.isdir(src_name):
            outputs.extend(safe_copy_tree(src_name, dst_name))

        else:
            safe_copy_file(src_name, dst_name)
            outputs.append(dst_name)

    return outputs



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"""Project description variables."""

PROJECT_NAME = "kapitan"
VERSION = "0.31.1rc3"
DESCRIPTION = "Generic templated configuration management for Kubernetes, Terraform and other things"
AUTHOR = "Ricardo Amaro"
AUTHOR_EMAIL = "ramaro@kapicorp.com"
LICENCE = "Apache License 2.0"
URL = "https://github.com/kapicorp/kapitan"



"azkms secret module"

import os
import logging
import base64
from urllib.parse import urlparse
from azure.keyvault.keys.crypto import CryptographyClient, EncryptionAlgorithm
from azure.keyvault.keys import KeyClient
from azure.identity import DefaultAzureCredential

from kapitan.refs.base64 import Base64Ref, Base64RefBackend
from kapitan.refs.base import RefError
from kapitan import cached
from kapitan.errors import KapitanError

logger = logging.getLogger(__name__)


class AzureKMSError(KapitanError):
    """
    Generic Azure Key Vault error
    """

    pass


def azkms_obj(key_id):
    """
    Return Azure Key Vault Object
    """
    # e.g of key_id https://kapitanbackend.vault.azure.net/keys/myKey/deadbeef
    if not cached.azkms_obj:
        url = urlparse(key_id)
        # ['', 'keys', 'myKey', 'deadbeef'] or ['kapitanbackend.vault.azure.net', 'keys', 'myKey', 'deadbeef']
        # depending on if key_id is prefixed with https://
        attrs = url.path.split("/")
        key_vault_uri = url.hostname or attrs[0]
        key_name = attrs[-2]
        key_version = attrs[-1]

        # If --verbose is set, show requests from azure
        if logger.getEffectiveLevel() > logging.DEBUG:
            logging.getLogger("azure").setLevel(logging.ERROR)
        credential = DefaultAzureCredential()
        key_client = KeyClient(vault_url=f"https://{key_vault_uri}", credential=credential)
        key = key_client.get_key(key_name, key_version)
        cached.azkms_obj = CryptographyClient(key, credential)

    return cached.azkms_obj


class AzureKMSSecret(Base64Ref):
    def __init__(self, data, key, encrypt=True, encode_base64=False, **kwargs):
        """
        encrypts data with key
        set encode_base64 to True to base64 encode data before encrypting and writing
        set encrypt to False if loading data that is already encrypted and base64
        """

        if encrypt:
            self._encrypt(data, key, encode_base64)
            if encode_base64:
                kwargs["encoding"] = "base64"
        else:
            self.data = data
            self.key = key
        super().__init__(self.data, **kwargs)
        self.type_name = "azkms"

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new AzureKMSSecret from data and ref_params: target_name
        key will be grabbed from the inventory via target_name
        """
        try:
            target_name = ref_params.kwargs["target_name"]
            if target_name is None:
                raise ValueError("target_name not set")

            target_inv = cached.inv["nodes"].get(target_name, None)
            if target_inv is None:
                raise ValueError("target_inv not set")

            key = target_inv["parameters"]["kapitan"]["secrets"]["azkms"]["key"]
            return cls(data, key, **ref_params.kwargs)
        except KeyError:
            raise RefError("Could not create AzureKMSSecret: target_name missing")

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        return super().from_path(ref_full_path, encrypt=False, **kwargs)

    def reveal(self):
        """
        returns decrypted data
        raises AzureKMSError if decryption fails
        """
        # can't use super().reveal() as we want bytes
        ref_data = base64.b64decode(self.data)
        return self._decrypt(ref_data, self.key)

    def update_key(self, key):
        """
        re-encrypts data with new key, respects original encoding
        returns True if key is different and secret is updated, False otherwise
        """
        if key == self.key:
            return False

        data_dec = self.reveal()
        encode_base64 = self.encoding == "base64"
        if encode_base64:
            data_dec = base64.b64decode(data_dec).decode()
        self._encrypt(data_dec, key, encode_base64)
        self.data = base64.b64encode(self.data).decode()
        return True

    def _encrypt(self, data, key, encode_base64):
        """
        encrypts data
        set encode_base64 to True to base64 encode data before writing
        """
        assert isinstance(key, str)
        _data = data
        self.encoding = "original"
        if encode_base64:
            _data = base64.b64encode(data.encode())
            self.encoding = "base64"
        else:
            # To guarantee _data is bytes
            if isinstance(data, str):
                _data = data.encode()
        try:
            ciphertext = ""
            # Mocking encrypted response for tests
            if key == "mock":
                ciphertext = base64.b64encode(_data)
            else:
                request = azkms_obj(key).encrypt(EncryptionAlgorithm.rsa_oaep_256, _data)
                ciphertext = request.ciphertext

            self.data = ciphertext
            self.key = key

        except Exception as e:
            raise AzureKMSError(e)

    def _decrypt(self, data, key):
        """decrypt data"""
        try:
            plaintext = ""
            # Mocking decrypted response for tests
            if self.key == "mock":
                plaintext = "mock".encode()
            else:
                request = azkms_obj(key).decrypt(EncryptionAlgorithm.rsa_oaep_256, data)
                plaintext = request.plaintext

            return plaintext.decode()

        except Exception as e:
            raise AzureKMSError(e)

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {"data": self.data, "encoding": self.encoding, "key": self.key, "type": self.type_name}


class AzureKMSBackend(Base64RefBackend):
    def __init__(self, path, ref_type=AzureKMSSecret, **ref_kwargs):
        "init AzureKMSBackend ref backend type"
        super().__init__(path, ref_type, **ref_kwargs)
        self.type_name = "azkms"



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"lint module"

import logging
import os
import sys
from pprint import pformat

from kapitan.errors import KapitanError
from kapitan.utils import list_all_paths
from yamllint import linter
from yamllint.config import YamlLintConfig

logger = logging.getLogger(__name__)

yamllint_config = """
# https://yamllint.readthedocs.io/en/stable/rules.html
rules:
  braces: disable
  brackets: disable
  colons: disable
  commas: disable
  comments: disable
  comments-indentation: disable
  document-end: disable
  document-start: disable
  empty-lines: disable
  empty-values: disable
  hyphens: disable
  indentation: disable
  key-duplicates: enable
  key-ordering: disable
  line-length: disable
  new-line-at-end-of-file: disable
  new-lines: disable
  octal-values: enable
  quoted-strings: disable
  trailing-spaces: disable
  truthy: disable
"""


def start_lint(args):
    """Runs all lint operations available
    Args:
        fail_on_warning (bool): if set to True, function will exit if any warning is found
        skip_class_checks (bool): whether to skip checking for class related warnings or not
        skip_yamllint (bool): whether to skip checking yaml files for lint problems
        inventory_path (string): path to your inventory/ folder
        search_secrets (bool): whether to search for secret related warnings or not
        secrets_path (string): path to your refs/ folder
        compiled_path (string): path to your compiled/ folder
    Yields:
        checks_sum (int): the number of lint warnings found
    """
    if args.skip_class_checks and args.skip_yamllint and not args.search_secrets:
        logger.info("Nothing to check. Remove --skip-class-checks or add --search-secrets to lint secrets")
        sys.exit(1)

    status_yamllint = 0
    status_secrets = 0
    status_class_checks = 0

    if not os.path.isdir(args.inventory_path):
        logger.info(
            "\nInventory path is invalid or not provided, skipping yamllint and orphan class checks\n"
        )
    else:
        if not args.skip_yamllint:
            logger.info("\nRunning yamllint on all inventory files...\n")
            status_yamllint = lint_yamllint(args.inventory_path)

        if not args.skip_class_checks:
            logger.info("\nChecking for orphan classes in inventory...\n")
            status_class_checks = lint_unused_classes(args.inventory_path)

    if args.search_secrets:
        logger.info("\nChecking for orphan secrets files...\n")
        status_secrets = lint_orphan_secrets(args.compiled_path, args.refs_path)

    checks_sum = status_secrets + status_class_checks + status_yamllint
    if args.fail_on_warning and checks_sum > 0:
        sys.exit(1)

    return checks_sum


def lint_orphan_secrets(compiled_path, secrets_path):
    """Checks your refs/ folder for unused secrets files by:
        - iterating the secrets_path/ dir and extracting all secrets names from the file paths
        - does a text search over the entire compiled_path/ to find usages of those secrets
    Args:
        compiled_path (string): path to your compiled/ folder
        secrets_path (string): path to your refs/ folder
    Yields:
        checks_sum (int): the number of orphan secrets found
    """
    logger.debug("Find secret paths for %s", secrets_path)
    secrets_paths = set()
    for path in list_all_paths(secrets_path):
        if os.path.isfile(path):
            path = path[len(secrets_path) + 1 :]
            secrets_paths.add(path)

    logger.debug("Collected # of paths: %s", len(secrets_paths))
    logger.debug("Checking if all secrets are declared in %s", compiled_path)

    for path in list_all_paths(compiled_path):
        if os.path.isfile(path):
            with open(path, "r") as compiled_file:
                file_contents = compiled_file.read()
                for secret_path in list(secrets_paths):
                    if secret_path in file_contents:
                        secrets_paths.discard(secret_path)

    checks_sum = len(secrets_paths)
    if checks_sum > 0:
        logger.info(
            "No usage found for the following %s secrets files:\n%s\n",
            len(secrets_paths),
            pformat(secrets_paths),
        )

    return checks_sum


def lint_unused_classes(inventory_path):
    """Checks your inventory for unused classes by:
        - iterating the inventory_path/classes/ dir and extracting all class names from the file paths
        - converting those file paths to class references (e.g. component/mysql -> component.mysql)
        - does a text search over the entire inventory_path/ to find usages of those classes
    Args:
        inventory_path (string): path to your inventory folder
    Yields:
        checks_sum (int): the number of unused classes found
    """
    classes_dir = os.path.join(inventory_path, "classes/")
    if not os.path.isdir(classes_dir):
        raise KapitanError("{} is not a valid directory or does not exist".format(classes_dir))

    logger.debug("Find unused classes from %s", classes_dir)
    class_paths = set()
    for path in list_all_paths(classes_dir):
        if os.path.isfile(path) and (path.endswith(".yml") or path.endswith(".yaml")):
            path = path[len(classes_dir) :]
            path = path.replace(".yml", "").replace(".yaml", "").replace("/", ".")
            class_paths.add(path)

    logger.debug("Collected # of paths: %s", len(class_paths))
    logger.debug("Checking if all classes are declared in %s", classes_dir)

    for path in list_all_paths(inventory_path):
        if os.path.isfile(path):
            with open(path, "r") as compiled_file:
                file_contents = compiled_file.read()
                for class_path in list(class_paths):
                    exists = class_path in file_contents
                    """
                    Classes files may reside in subdirectories, which act as namespaces.
                    For instance, a class ssh.server will result in the class definition to be read from ssh/server.yml.
                    Specifying just ssh will cause the class data to be read from ssh/init.yml or ssh.yml.
                    Note, however, that only one of those two may be present.
                    Thus we also check for ".init" being used in the class here to cover that case.
                    https://reclass.pantsfullofunix.net/operations.html
                    """
                    if class_path.endswith(".init"):
                        exists = (class_path[:-5] in file_contents) or (exists)

                    if exists:
                        class_paths.discard(class_path)

    checks_sum = len(class_paths)
    if checks_sum > 0:
        logger.info(
            "No usage found for the following %s classes:\n%s\n", len(class_paths), pformat(class_paths)
        )

    return checks_sum


def lint_yamllint(inventory_path):
    """Run yamllint on all yaml files in inventory
    Args:
        inventory_path (string): path to your inventory/ folder
    Yields:
        checks_sum (int): the number of yaml lint issues found
    """
    logger.debug("Running yamllint for %s", inventory_path)

    if os.path.isfile(".yamllint"):
        logger.info("Loading values from .yamllint found.")
        conf = YamlLintConfig(file=".yamllint")
    else:
        logger.info(".yamllint not found. Using default values")
        conf = YamlLintConfig(yamllint_config)

    checks_sum = 0
    for path in list_all_paths(inventory_path):
        if os.path.isfile(path) and (path.endswith(".yml") or path.endswith(".yaml")):
            with open(path, "r") as yaml_file:
                file_contents = yaml_file.read()

                try:
                    problems = list(linter.run(file_contents, conf, filepath=path))
                except EnvironmentError as e:
                    logger.error(e)
                    sys.exit(-1)

                if len(problems) > 0:
                    checks_sum += len(problems)
                    logger.info("File %s has the following issues:", path)
                    for problem in problems:
                        logger.info("\t%s", problem)

    if checks_sum > 0:
        logger.info("\nTotal yamllint issues found: %s", checks_sum)

    return checks_sum



# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging

logger = logging.getLogger(__name__)


class Validator(object):
    def __init__(self, cache_dir, **kwargs):
        self.cache_dir = cache_dir

    def validate(self, validate_obj, **kwargs):
        raise NotImplementedError



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pyparsing as pp

from .compitem import CompItem
from .invitem import InvItem
from .refitem import RefItem
from .scaitem import ScaItem

from reclass.errors import ParseError
from reclass.values.parser_funcs import tags
import reclass.values.parser_funcs as parsers

import collections
import six


class Parser(object):

    def __init__(self):
        self._ref_parser = None
        self._simple_parser = None
        self._old_settings = None

    @property
    def ref_parser(self):
        if self._ref_parser is None or self._settings != self._old_settings:
            self._ref_parser = parsers.get_ref_parser(self._settings)
            self._old_settings = self._settings
        return self._ref_parser

    @property
    def simple_ref_parser(self):
        if self._simple_parser is None or self._settings != self._old_settings:
            self._simple_parser = parsers.get_simple_ref_parser(self._settings)
            self._old_settings = self._settings
        return self._simple_parser

    def parse(self, value, settings):
        def full_parse():
            try:
                return self.ref_parser.parseString(value)
            except pp.ParseException as e:
                raise ParseError(e.msg, e.line, e.col, e.lineno)

        self._settings = settings
        sentinel_count = (value.count(settings.reference_sentinels[0]) +
                          value.count(settings.export_sentinels[0]))
        if sentinel_count == 0:
            # speed up: only use pyparsing if there are sentinels in the value
            return ScaItem(value, self._settings)
        elif sentinel_count == 1:  # speed up: try a simple reference
            try:
                tokens = self.simple_ref_parser.parseString(value)
            except pp.ParseException:
                tokens = full_parse()  # fall back on the full parser
        else:
            tokens = full_parse()  # use the full parser

        tokens = parsers.listify(tokens)
        items = self._create_items(tokens)
        if len(items) == 1:
            return items[0]
        return CompItem(items, self._settings)

    _item_builders = {tags.STR: (lambda s, v: ScaItem(v, s._settings)),
                      tags.REF: (lambda s, v: s._create_ref(v)),
                      tags.INV: (lambda s, v: s._create_inv(v)) }

    def _create_items(self, tokens):
        return [self._item_builders[t](self, v) for t, v in tokens ]

    def _create_ref(self, tokens):
        items = [ self._item_builders[t](self, v) for t, v in tokens ]
        return RefItem(items, self._settings)

    def _create_inv(self, tokens):
        items = [ScaItem(v, self._settings) for t, v in tokens]
        if len(items) == 1:
            return InvItem(items[0], self._settings)
        return InvItem(CompItem(items), self._settings)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


from .classes import Classes

class Applications(Classes):
    '''
    Extends Classes with the possibility to let specially formatted items
    remove earlier occurences of the item. For instance, if the "negater" is
    '~', then "adding" an element "~foo" to a list causes a previous element
    "foo" to be removed. If no such element exists, nothing happens, but
    a reference of the negation is kept, in case the instance is later used to
    extend another instance, in which case the negations should apply to the
    instance to be extended.
    '''
    DEFAULT_NEGATION_PREFIX = '~'

    def __init__(self, iterable=None,
                 negation_prefix=DEFAULT_NEGATION_PREFIX):
        self.negation_prefix = negation_prefix
        self._offset = len(negation_prefix)
        self._negations = []
        super(Applications, self).__init__(iterable)

    def append_if_new(self, item):
        self._assert_is_string(item)
        if item.startswith(self.negation_prefix):
            item = item[self._offset:]
            self._negations.append(item)
            try:
                self._items.remove(item)
            except ValueError:
                pass
        else:
            super(Applications, self)._append_if_new(item)

    def merge_unique(self, iterable):
        if isinstance(iterable, self.__class__):
            # we might be extending ourselves to include negated applications,
            # in which case we need to remove our own content accordingly:
            for negation in iterable._negations:
                try:
                    self._items.remove(negation)
                except ValueError:
                    pass
            iterable = iterable.as_list()
        for i in iterable:
            self.append_if_new(i)

    def __repr__(self):
        contents = self._items + \
                ['%s%s' % (self.negation_prefix, i) for i in self._negations]
        return "%s(%r, %r)" % (self.__class__.__name__, contents,
                               str(self.negation_prefix))



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"initialiser module"

import logging
import os
from distutils.dir_util import copy_tree

logger = logging.getLogger(__name__)


def initialise_skeleton(args):
    """Initialises a directory with a recommended skeleton structure
    Args:
        args.directory (string): path which to initialise, directory is assumed to exist
    """

    current_pwd = os.path.dirname(__file__)
    templates_directory = os.path.join(current_pwd, "inputs", "templates")
    populated = copy_tree(templates_directory, args.directory)
    logger.info("Populated %s with:", args.directory)
    for directory, subs, file_list in os.walk(args.directory):
        # In order to avoid adding the given directory itself in listing.
        if directory == args.directory:
            continue
        if any([path.startswith(directory) for path in populated]):
            level = directory.replace(args.directory, "").count(os.sep) - 1
            indent = " " * 4 * (level)
            logger.info("%s%s", indent, os.path.basename(directory))
            for fname in file_list:
                if os.path.join(directory, fname) in populated:
                    sub_indent = " " * 4 * (level + 1)
                    logger.info("%s%s", sub_indent, fname)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import time
import re
import fnmatch
import shlex
import os
import sys
import yaml

from six import iteritems

from reclass.settings import Settings
from reclass.datatypes import Entity, Classes, Parameters, Exports
from reclass.errors import MappingFormatError, ClassNameResolveError, ClassNotFound, InvQueryClassNameResolveError, InvQueryClassNotFound, InvQueryError, InterpolationError, ResolveError
from reclass.values.parser import Parser


class Core(object):

    _parser = Parser()

    def __init__(self, storage, class_mappings, settings, input_data=None):
        self._storage = storage
        self._class_mappings = class_mappings
        self._settings = settings
        self._input_data = input_data
        if self._settings.ignore_class_notfound:
            self._cnf_r = re.compile(
                '|'.join(self._settings.ignore_class_notfound_regexp))

    @staticmethod
    def _get_timestamp():
        return time.strftime('%c')

    @staticmethod
    def _match_regexp(key, nodename):
        return re.search(key, nodename)

    @staticmethod
    def _match_glob(key, nodename):
        return fnmatch.fnmatchcase(nodename, key)

    @staticmethod
    def _shlex_split(instr):
        lexer = shlex.shlex(instr, posix=True)
        lexer.whitespace_split = True
        lexer.commenters = ''
        regexp = False
        if instr[0] == '/':
            lexer.quotes += '/'
            lexer.escapedquotes += '/'
            regexp = True
        try:
            key = lexer.get_token()
        except ValueError as e:
            raise MappingFormatError('Error in mapping "{0}": missing closing '
                                     'quote (or slash)'.format(instr))
        if regexp:
            key = '/{0}/'.format(key)
        return key, list(lexer)

    def _get_class_mappings_entity(self, entity):
        if not self._class_mappings:
            return Entity(self._settings, name='empty (class mappings)')
        c = Classes()
        if self._settings.class_mappings_match_path:
            matchname = entity.pathname
        else:
            matchname = entity.name
        for mapping in self._class_mappings:
            matched = False
            key, klasses = Core._shlex_split(mapping)
            if key[0] == ('/'):
                matched = Core._match_regexp(key[1:-1], matchname)
                if matched:
                    for klass in klasses:
                        c.append_if_new(matched.expand(klass))

            else:
                if Core._match_glob(key, matchname):
                    for klass in klasses:
                        c.append_if_new(klass)

        return Entity(self._settings, classes=c,
                      name='class mappings for node {0}'.format(entity.name))

    def _get_input_data_entity(self):
        if not self._input_data:
            return Entity(self._settings, name='empty (input data)')
        p = Parameters(self._input_data, self._settings)
        return Entity(self._settings, parameters=p, name='input data')

    def _recurse_entity(self, entity, merge_base=None, seen=None, nodename=None, environment=None):
        if seen is None:
            seen = {}

        if environment is None:
            environment = self._settings.default_environment

        if merge_base is None:
            merge_base = Entity(self._settings, name='empty (@{0})'.format(nodename))

        for klass in entity.classes.as_list():
            # class name contain reference
            num_references = klass.count(self._settings.reference_sentinels[0]) +\
                             klass.count(self._settings.export_sentinels[0])
            if num_references > 0:
                # Make a copy of merge_base.parameters to avoid running into
                # issues as new data is merged into merge_base.parameters.
                # We need to make a new copy on each pass to ensure we don't
                # try to resolve references using a stale set of parameters,
                # as merge_base is updated in the loop.
                mbparams = copy.deepcopy(merge_base.parameters)
                # Enable interpolation for copy of parameters before using the
                # copy to resolve references in class names
                mbparams.initialise_interpolation()
                try:
                    klass = str(self._parser.parse(klass, self._settings).render(mbparams.as_dict(), {}))
                except ResolveError as e:
                    raise ClassNameResolveError(klass, nodename, entity.uri)

            if klass not in seen:
                try:
                    class_entity = self._storage.get_class(klass, environment, self._settings)
                except ClassNotFound as e:
                    if self._settings.ignore_class_notfound:
                        if self._cnf_r.match(klass):
                            if self._settings.ignore_class_notfound_warning:
                                # TODO, add logging handler
                                print("[WARNING] Reclass class not found: '%s'. Skipped!" % klass, file=sys.stderr)
                            continue
                    e.nodename = nodename
                    e.uri = entity.uri
                    raise

                # on every iteration, we pass what we have so far into the
                # recursive descent …
                descent = self._recurse_entity(class_entity, merge_base=merge_base, seen=seen,
                                               nodename=nodename, environment=environment)
                # … therefore, we don't need to merge the result of the
                # recursive descent as the result is a reference to the same
                # merge_base object we passed to the call originally, with the
                # new values merged in …
                assert descent == merge_base
                seen[klass] = True

        # … and finally, we merge what we have at this level into the
        # result of the iteration, so that elements at the current level
        # overwrite stuff defined by parents
        merge_base.merge(entity)
        return merge_base

    def _get_automatic_parameters(self, nodename, environment):
        if self._settings.automatic_parameters:
            pars = {
                '_reclass_': {
                    'name': {
                        'full': nodename,
                        'short': nodename.split('.').pop(),
                        'path': os.path.join(*nodename.split('.')),
                        'parts': nodename.split('.')
                    },
                'environment': environment
                }
            }
            return Parameters(pars, self._settings, '__auto__')
        else:
            return Parameters({}, self._settings, '')

    def _get_inventory(self, all_envs, environment, queries):
        inventory = {}
        for nodename in self._storage.enumerate_nodes():
            try:
                node_base = self._storage.get_node(nodename, self._settings)
                if node_base.environment is None:
                    node_base.environment = self._settings.default_environment
            except yaml.scanner.ScannerError as e:
                if self._settings.inventory_ignore_failed_node:
                    continue
                raise

            if all_envs or node_base.environment == environment:
                try:
                    node = self._node_entity(nodename)
                except ClassNotFound as e:
                    raise InvQueryClassNotFound(e)
                except ClassNameResolveError as e:
                    raise InvQueryClassNameResolveError(e)
                if queries is None:
                    try:
                        node.interpolate_exports()
                    except InterpolationError as e:
                        e.nodename = nodename
                else:
                    node.initialise_interpolation()
                    for p, q in queries:
                        try:
                            node.interpolate_single_export(q)
                        except InterpolationError as e:
                            e.nodename = nodename
                            raise InvQueryError(q.contents, e, context=p, uri=q.uri)
                inventory[nodename] = node.exports.as_dict()
        return inventory

    def _node_entity(self, nodename):
        node_entity = self._storage.get_node(nodename, self._settings)
        if node_entity.environment == None:
            node_entity.environment = self._settings.default_environment
        base_entity = Entity(self._settings, name='base')
        base_entity.merge(self._get_class_mappings_entity(node_entity))
        base_entity.merge(self._get_input_data_entity())
        base_entity.merge_parameters(self._get_automatic_parameters(nodename, node_entity.environment))
        seen = {}
        merge_base = self._recurse_entity(base_entity, seen=seen, nodename=nodename,
                                          environment=node_entity.environment)
        return self._recurse_entity(node_entity, merge_base=merge_base, seen=seen,
                                    nodename=nodename, environment=node_entity.environment)

    def _nodeinfo(self, nodename, inventory):
        try:
            node = self._node_entity(nodename)
            node.initialise_interpolation()
            if node.parameters.has_inv_query and inventory is None:
                inventory = self._get_inventory(node.parameters.needs_all_envs, node.environment, node.parameters.get_inv_queries())
            node.interpolate(inventory)
            return node
        except InterpolationError as e:
            e.nodename = nodename
            raise

    def _nodeinfo_as_dict(self, nodename, entity):
        ret = {'__reclass__' : {'node': entity.name,
                                'name': nodename,
                                'uri': entity.uri,
                                'environment': entity.environment,
                                'timestamp': Core._get_timestamp()
                               },
              }
        ret.update(entity.as_dict())
        return ret

    def nodeinfo(self, nodename):
        return self._nodeinfo_as_dict(nodename, self._nodeinfo(nodename, None))

    def inventory(self):
        query_nodes = set()
        entities = {}
        inventory = self._get_inventory(True, '', None)
        for n in self._storage.enumerate_nodes():
            entities[n] = self._nodeinfo(n, inventory)
            if entities[n].parameters.has_inv_query:
                nodes.add(n)
        for n in query_nodes:
            entities[n] = self._nodeinfo(n, inventory)

        nodes = {}
        applications = {}
        classes = {}
        for (f, nodeinfo) in iteritems(entities):
            d = nodes[f] = self._nodeinfo_as_dict(f, nodeinfo)
            for a in d['applications']:
                if a in applications:
                    applications[a].append(f)
                else:
                    applications[a] = [f]
            for c in d['classes']:
                if c in classes:
                    classes[c].append(f)
                else:
                    classes[c] = [f]

        return {'__reclass__' : {'timestamp': Core._get_timestamp()},
                'nodes': nodes,
                'classes': classes,
                'applications': applications
               }



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.utils.dictpath import DictPath
import unittest

class TestDictPath(unittest.TestCase):

    def test_constructor0(self):
        p = DictPath(':')
        self.assertListEqual(p._parts, [])

    def test_constructor_list(self):
        l = ['a', 'b', 'c']
        p = DictPath(':', l)
        self.assertListEqual(p._parts, l)

    def test_constructor_str(self):
        delim = ':'
        s = 'a{0}b{0}c'.format(delim)
        l = ['a', 'b', 'c']
        p = DictPath(delim, s)
        self.assertListEqual(p._parts, l)

    def test_constructor_str_escaped(self):
        delim = ':'
        s = 'a{0}b\{0}b{0}c'.format(delim)
        l = ['a', 'b\\{0}b'.format(delim), 'c']
        p = DictPath(delim, s)
        self.assertListEqual(p._parts, l)

    def test_constructor_invalid_type(self):
        with self.assertRaises(TypeError):
            p = DictPath(':', 5)

    def test_equality(self):
        delim = ':'
        s = 'a{0}b{0}c'.format(delim)
        l = ['a', 'b', 'c']
        p1 = DictPath(delim, s)
        p2 = DictPath(delim, l)
        self.assertEqual(p1, p2)

    def test_inequality_content(self):
        delim = ':'
        s = 'a{0}b{0}c'.format(delim)
        l = ['d', 'e', 'f']
        p1 = DictPath(delim, s)
        p2 = DictPath(delim, l)
        self.assertNotEqual(p1, p2)

    def test_inequality_delimiter(self):
        l = ['a', 'b', 'c']
        p1 = DictPath(':', l)
        p2 = DictPath('%', l)
        self.assertNotEqual(p1, p2)

    def test_repr(self):
        delim = '%'
        s = 'a:b\:b:c'
        p = DictPath(delim, s)
        self.assertEqual('%r' % p, "DictPath(%r, %r)" % (delim, str(s)))

    def test_str(self):
        s = 'a:b\:b:c'
        p = DictPath(':', s)
        self.assertEqual(str(p), s)

    def test_path_accessor(self):
        l = ['a', 'b', 'c']
        p = DictPath(':', l)
        self.assertListEqual(p.path, l)

    def test_new_subpath(self):
        l = ['a', 'b', 'c']
        p = DictPath(':', l[:-1])
        p = p.new_subpath(l[-1])
        self.assertListEqual(p.path, l)

    def test_get_value(self):
        v = 42
        l = ['a', 'b', 'c']
        d = {'a':{'b':{'c':v}}}
        p = DictPath(':', l)
        self.assertEqual(p.get_value(d), v)

    def test_get_value_escaped(self):
        v = 42
        l = ['a', 'b:b', 'c']
        d = {'a':{'b:b':{'c':v}}}
        p = DictPath(':', l)
        self.assertEqual(p.get_value(d), v)

    def test_get_value_listindex_list(self):
        v = 42
        l = ['a', 1, 'c']
        d = {'a':[None, {'c':v}, None]}
        p = DictPath(':', l)
        self.assertEqual(p.get_value(d), v)

    def test_get_value_listindex_str(self):
        v = 42
        s = 'a:1:c'
        d = {'a':[None, {'c':v}, None]}
        p = DictPath(':', s)
        self.assertEqual(p.get_value(d), v)

    def test_set_value(self):
        v = 42
        l = ['a', 'b', 'c']
        d = {'a':{'b':{'c':v}}}
        p = DictPath(':', l)
        p.set_value(d, v+1)
        self.assertEqual(d['a']['b']['c'], v+1)

    def test_set_value_escaped(self):
        v = 42
        l = ['a', 'b:b', 'c']
        d = {'a':{'b:b':{'c':v}}}
        p = DictPath(':', l)
        p.set_value(d, v+1)
        self.assertEqual(d['a']['b:b']['c'], v+1)

    def test_set_value_escaped_listindex_list(self):
        v = 42
        l = ['a', 1, 'c']
        d = {'a':[None, {'c':v}, None]}
        p = DictPath(':', l)
        p.set_value(d, v+1)
        self.assertEqual(d['a'][1]['c'], v+1)

    def test_set_value_escaped_listindex_str(self):
        v = 42
        s = 'a:1:c'
        d = {'a':[None, {'c':v}, None]}
        p = DictPath(':', s)
        p.set_value(d, v+1)
        self.assertEqual(d['a'][1]['c'], v+1)

    def test_get_nonexistent_value(self):
        l = ['a', 'd']
        p = DictPath(':', l)
        with self.assertRaises(KeyError):
            p.get_value(dict())

    def test_set_nonexistent_value(self):
        l = ['a', 'd']
        p = DictPath(':', l)
        with self.assertRaises(KeyError):
            p.set_value(dict(), 42)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import subprocess
import re
import logging
import os
import shutil
from distutils.dir_util import copy_tree

from kapitan.inputs.base import InputType

logger = logging.getLogger(__name__)


class External(InputType):
    def __init__(self, compile_path, search_paths, ref_controller):
        super().__init__("external", compile_path, search_paths, ref_controller)
        self.env_vars = {}
        self.args = []
        self.target_name = None

    def set_env_vars(self, env_vars):
        # Propagate HOME and PATH environment variables to external tool
        # This is necessary, because calling `subprocess.run()` with `env` set doesn't propagate
        # any environment variables from the current process.
        # We only propagate HOME or PATH if they're present in Kapitan's environment, but aren't
        # explicitly specified in the input's env_vars already. This ensures we don't run into
        # issues when spawning the subprocess due to `None` values being present in the subprocess's
        # environment.
        if "PATH" not in env_vars and "PATH" in os.environ:
            env_vars["PATH"] = os.environ["PATH"]
        if "HOME" not in env_vars and "HOME" in os.environ:
            env_vars["HOME"] = os.environ["HOME"]
        self.env_vars = env_vars

    def set_args(self, args):
        self.args = args

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Execute external with specific arguments and env vars.
        If external exits with non zero error code, error is thrown
        """

        try:
            # file_path (str): Path to executable script or binary
            external_path = file_path

            args = [external_path]
            args.extend(self.args)
            args = " ".join(args)

            # compile_path (str): Path to current target compiled directory
            compiled_target_pattern = re.compile(r"(\${compiled_target_dir})")
            args = compiled_target_pattern.sub(compile_path, args)
            # substitute `${compiled_target_dir}` in provided environment variables
            env_vars = {k: compiled_target_pattern.sub(compile_path, v) for (k, v) in self.env_vars.items()}

            logger.debug("Executing external input with command '%s' and env vars '%s'.", args, env_vars)

            external_result = subprocess.run(
                args,
                env=env_vars,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
                encoding="utf8",
            )

            logger.debug("External stdout: %s.", external_result.stdout)
            if external_result.returncode != 0:
                raise ValueError(
                    "Executing external input with command '{}' and env vars '{}' failed: {}".format(
                        args, env_vars, external_result.stderr
                    )
                )

        except OSError as e:
            logger.exception("External failed to run. Error: %s", e)

    def default_output_type(self):
        # no output_type options for external
        return None



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import os
import sys
import logging


def setup_logging(name=None, level=logging.INFO, force=False):
    "setup logging and deal with logging behaviours in MacOS python 3.8 and below"
    # default opts
    kwopts = {"format": "%(message)s", "level": level}

    if level == logging.DEBUG:
        kwopts["format"] = "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"

    if sys.version_info >= (3, 8) and force:
        kwopts["force"] = True

    logging.basicConfig(**kwopts)

    if sys.version_info < (3, 8) and force:
        logging.getLogger(name).setLevel(level)


# XXX in MacOS, updating logging level in __main__ doesn't work for python3.8+
# XXX this is a hack that seems to work
if "-v" in sys.argv or "--verbose" in sys.argv:
    setup_logging(level=logging.DEBUG)
else:
    setup_logging()

# Adding reclass to PYTHONPATH
sys.path.insert(0, os.path.dirname(__file__) + "/reclass")



import hashlib
import logging
import os
from collections import defaultdict
from functools import partial

from kapitan import cached
from kapitan.dependency_manager.base import fetch_git_dependency, fetch_http_dependency
from kapitan.utils import normalise_join_path

logger = logging.getLogger(__name__)


def fetch_inventories(inventory_path, target_objs, save_dir, force, pool):
    """Parses through the 'inventory' parameter in target_objs to fetch the remote
    inventories and stores it in save_dir before recursively
    copying it to the output_path (relative to the inventory path)
    Overwrites existing inventory items if force fetched

    :param inventory_path: default or user specified inventory path
    :param target_objs: target objects
    :param save_dir: directory to save the fetched items
    :param force: bool value
    :param pool: pool object for multiprocessing
    :return: None
    """
    git_inventories = defaultdict(list)
    http_inventories = defaultdict(list)
    # To ensure no duplicate output path
    inv_output_path = defaultdict(set)

    for target_obj in target_objs:
        try:
            inventories = target_obj["inventory"]
            for inv in inventories:
                inv_type = inv["type"]
                source_uri = inv["source"]
                source_hash = hashlib.sha256(source_uri.encode())
                # hashing the source, subdir and ref together for git sources
                # as different inventory items can have the same git uri
                if "subdir" in inv:
                    subdir = inv["subdir"]
                    source_hash.update(subdir.encode())
                if "ref" in inv:
                    ref = inv["ref"]
                    source_hash.update(ref.encode())
                if source_hash in cached.inv_sources:
                    continue
                # output_path is relative to inventory_path
                # ./inventory by default
                output_path = normalise_join_path(inventory_path, inv["output_path"])
                logger.debug("Updated output_path from %s to %s", inv["output_path"], output_path)
                inv["output_path"] = output_path

                if output_path in inv_output_path[source_uri]:
                    # if the output_path is duplicated for the same source_uri
                    logger.warning("Skipping duplicate output path for uri %s", source_uri)
                    continue
                else:
                    inv_output_path[source_uri].add(output_path)

                if inv_type == "git":
                    git_inventories[source_uri].append(inv)
                elif inv_type in ("http", "https"):
                    http_inventories[source_uri].append(inv)
                else:
                    logger.warning("%s is not a valid source type", inv_type)
        except KeyError:
            logger.debug("Target object %s has no inventory key", target_obj["vars"]["target"])
            continue

    git_worker = partial(fetch_git_dependency, save_dir=save_dir, force=force, item_type="Inventory")
    http_worker = partial(fetch_http_dependency, save_dir=save_dir, force=force, item_type="Inventory")
    [p.get() for p in pool.imap_unordered(git_worker, git_inventories.items()) if p]
    [p.get() for p in pool.imap_unordered(http_worker, http_inventories.items()) if p]


def list_sources(target_objs):
    "returns list of all remote inventory sources"
    sources = []
    for target_obj in target_objs:
        try:
            invs = target_obj["inventory"]
            for inv in invs:
                source_uri = inv["source"]
                source_hash = hashlib.sha256(source_uri.encode())
                # hashing the source, subdir, ref together to idenntify unique items
                if "subdir" in inv:
                    subdir = inv["subdir"]
                    source_hash.update(subdir.encode())
                if "ref" in inv:
                    ref = inv["ref"]
                    source_hash.update(ref.encode())

                sources.append(source_hash.hexdigest()[:8])
        except KeyError:
            logger.debug(
                "Listing sources: target object %s has no 'inventory' key, continuing",
                target_obj["vars"]["target"],
            )
            continue
    return sources



import logging
import os
import subprocess
from subprocess import DEVNULL, PIPE

logger = logging.getLogger(__name__)


def helm_cli(helm_path, args, stdout=None, verbose=False):
    # if helm is not specified, try to get it from env var, and defaults to looking up helm in the path.
    if not helm_path:
        helm_path = os.getenv("KAPITAN_HELM_PATH", "helm")
    try:
        logger.debug("launching helm with arguments: %s", args)
        res = subprocess.run(
            args=[helm_path] + args, stderr=PIPE, stdout=stdout or (PIPE if verbose else DEVNULL)
        )
        if verbose and not stdout:
            for line in res.stdout.splitlines():
                if line:
                    logger.debug("[helm] %s", line.decode())
        return res.stderr.decode() if res.returncode != 0 else ""
    except FileNotFoundError:
        return "helm binary not found. helm must be present in the PATH to use kapitan helm functionalities"



# -*- coding: utf-8

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import reclass.defaults as defaults

from six import string_types, iteritems


class Settings(object):

    known_opts = {
        'allow_scalar_over_dict': defaults.OPT_ALLOW_SCALAR_OVER_DICT,
        'allow_scalar_over_list': defaults.OPT_ALLOW_SCALAR_OVER_LIST,
        'allow_list_over_scalar': defaults.OPT_ALLOW_LIST_OVER_SCALAR,
        'allow_dict_over_scalar': defaults.OPT_ALLOW_DICT_OVER_SCALAR,
        'allow_none_override': defaults.OPT_ALLOW_NONE_OVERRIDE,
        'automatic_parameters': defaults.AUTOMATIC_RECLASS_PARAMETERS,
        'class_mappings_match_path': defaults.CLASS_MAPPINGS_MATCH_PATH,
        'default_environment': defaults.DEFAULT_ENVIRONMENT,
        'delimiter': defaults.PARAMETER_INTERPOLATION_DELIMITER,
        'dict_key_override_prefix':
            defaults.PARAMETER_DICT_KEY_OVERRIDE_PREFIX,
        'dict_key_constant_prefix':
            defaults.PARAMETER_DICT_KEY_CONSTANT_PREFIX,
        'escape_character': defaults.ESCAPE_CHARACTER,
        'export_sentinels': defaults.EXPORT_SENTINELS,
        'inventory_ignore_failed_node':
            defaults.OPT_INVENTORY_IGNORE_FAILED_NODE,
        'inventory_ignore_failed_render':
            defaults.OPT_INVENTORY_IGNORE_FAILED_RENDER,
        'reference_sentinels': defaults.REFERENCE_SENTINELS,
        'ignore_class_notfound': defaults.OPT_IGNORE_CLASS_NOTFOUND,
        'strict_constant_parameters':
            defaults.OPT_STRICT_CONSTANT_PARAMETERS,
        'ignore_class_notfound_regexp':
            defaults.OPT_IGNORE_CLASS_NOTFOUND_REGEXP,
        'ignore_class_notfound_warning':
            defaults.OPT_IGNORE_CLASS_NOTFOUND_WARNING,
        'ignore_overwritten_missing_references':
            defaults.OPT_IGNORE_OVERWRITTEN_MISSING_REFERENCES,
        'group_errors': defaults.OPT_GROUP_ERRORS,
        'compose_node_name': defaults.OPT_COMPOSE_NODE_NAME,
    }

    def __init__(self, options={}):
        for opt_name, opt_value in iteritems(self.known_opts):
            setattr(self, opt_name, options.get(opt_name, opt_value))

        self.dict_key_prefixes = [str(self.dict_key_override_prefix),
                                  str(self.dict_key_constant_prefix)]
        if isinstance(self.ignore_class_notfound_regexp, string_types):
            self.ignore_class_notfound_regexp = [
                    self.ignore_class_notfound_regexp]

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return all(getattr(self, opt) == getattr(other, opt)
                       for opt in self.known_opts)
        return False

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        return self.__copy__()



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from enum import Enum

from reclass.utils.dictpath import DictPath

ItemTypes = Enum('ItemTypes',
                 ['COMPOSITE', 'DICTIONARY', 'INV_QUERY', 'LIST',
                  'REFERENCE', 'SCALAR'])


class Item(object):

    def __init__(self, item, settings):
        self._settings = settings
        self.contents = item
        self.has_inv_query = False

    def allRefs(self):
        return True

    @property
    def has_references(self):
        return False

    def is_container(self):
        return False

    @property
    def is_complex(self):
        return (self.has_references | self.has_inv_query)

    def merge_over(self, item):
        msg = "Item class {0} does not implement merge_over()"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def render(self, context, exports):
        msg = "Item class {0} does not implement render()"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def type_str(self):
        return self.type.name.lower()

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.contents)


class ItemWithReferences(Item):

    def __init__(self, items, settings):
        super(ItemWithReferences, self).__init__(items, settings)
        try:
            iter(self.contents)
        except TypeError:
            self.contents = [self.contents]
        self.assembleRefs()

    @property
    def has_references(self):
        return len(self._refs) > 0

    def get_references(self):
        return self._refs

    # NOTE: possibility of confusion. Looks like 'assemble' should be either
    # 'gather' or 'extract'.
    def assembleRefs(self, context={}):
        self._refs = []
        self.allRefs = True
        for item in self.contents:
            if item.has_references:
                item.assembleRefs(context)
                self._refs.extend(item.get_references())
                if item.allRefs is False:
                    self.allRefs = False


class ContainerItem(Item):

    def is_container(self):
        return True

    def render(self, context, inventory):
        return self.contents
